from collections import namedtuple

from dagster import check
from dagster.core.definitions import PipelineDefinition, ModeDefinition, SystemStorageDefinition
from dagster.core.execution.config import RunConfig
from dagster.core.system_config.objects import EnvironmentConfig
from dagster.core.storage.type_storage import TypeStoragePluginRegistry


class InitSystemStorageContext(
    namedtuple(
        'InitSystemStorageContext',
        (
            'pipeline_def mode_def system_storage_def run_config environment_config '
            'type_storage_plugin_registry resources system_storage_config'
        ),
    )
):
    def __new__(
        cls,
        pipeline_def,
        mode_def,
        system_storage_def,
        run_config,
        environment_config,
        type_storage_plugin_registry,
        resources,
        system_storage_config,
    ):
        return super(InitSystemStorageContext, cls).__new__(
            cls,
            pipeline_def=check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition),
            mode_def=check.inst_param(mode_def, 'mode_def', ModeDefinition),
            system_storage_def=check.inst_param(
                system_storage_def, 'system_storage_def', SystemStorageDefinition
            ),
            run_config=check.inst_param(run_config, 'run_config', RunConfig),
            environment_config=check.inst_param(
                environment_config, 'environment_config', EnvironmentConfig
            ),
            type_storage_plugin_registry=check.inst_param(
                type_storage_plugin_registry,
                'type_storage_plugin_registry',
                TypeStoragePluginRegistry,
            ),
            resources=check.not_none_param(resources, 'resources'),
            system_storage_config=check.dict_param(
                system_storage_config, system_storage_config, key_type=str
            ),
        )



'''Internal py2/3 compatibility library. A little more than six.'''

import os
import sys
import time

from .json import dump, dumps, JSONDecodeError
from .temp_dir import get_system_temp_directory

try:
    FileNotFoundError = FileNotFoundError  # pylint:disable=redefined-builtin
except NameError:
    FileNotFoundError = IOError

if sys.version_info.major >= 3:
    from io import StringIO  # pylint:disable=import-error
else:
    from StringIO import StringIO  # pylint:disable=import-error


# TODO implement a generic import by name -- see https://stackoverflow.com/questions/301134/how-to-import-a-module-given-its-name

# https://stackoverflow.com/a/67692/324449
def import_module_from_path(module_name, path_to_file):
    version = sys.version_info
    if version.major >= 3 and version.minor >= 5:
        import importlib.util

        spec = importlib.util.spec_from_file_location(module_name, path_to_file)
        module = importlib.util.module_from_spec(spec)
        sys.modules[spec.name] = module
        spec.loader.exec_module(module)
    elif version.major >= 3 and version.minor >= 3:
        from importlib.machinery import SourceFileLoader

        # pylint:disable=deprecated-method, no-value-for-parameter
        module = SourceFileLoader(module_name, path_to_file).load_module()
    else:
        from imp import load_source

        module = load_source(module_name, path_to_file)

    return module


# https://stackoverflow.com/a/437591/324449
def reload_module(module):
    version = sys.version_info
    if version.major >= 3 and version.minor >= 4:
        from importlib import reload as reload_

        return reload_(module)
    elif version.major >= 3:
        from imp import reload as reload_

        return reload_(module)

    return reload(module)  # pylint: disable=undefined-variable


def is_ascii(str_):
    if sys.version_info.major < 3:
        try:
            str_.decode('ascii')
            return True
        except UnicodeEncodeError:
            return False
    elif sys.version_info.major == 3 and sys.version_info.minor < 7:
        try:
            str_.encode('ascii')
            return True
        except UnicodeEncodeError:
            return False
    else:
        return str_.isascii()


if sys.version_info.major >= 3 and sys.version_info.minor >= 3:
    time_fn = time.perf_counter
elif os.name == 'nt':
    time_fn = time.clock
else:
    time_fn = time.time

try:
    from unittest import mock
except ImportError:
    # Because this dependency is not encoded setup.py deliberately
    # (we do not want to override or conflict with our users mocks)
    # we never fail when importing this.

    # This will only be used within *our* test enviroment of which
    # we have total control
    try:
        import mock
    except ImportError:
        pass



import pytest

from dagster.core.execution.api import create_execution_plan
from dagster.core.execution.context_creation_pipeline import DagsterInvalidConfigError
from ..engine_tests.test_multiprocessing import define_diamond_pipeline


def test_topological_sort():
    plan = create_execution_plan(define_diamond_pipeline())

    levels = plan.topological_step_levels()

    assert len(levels) == 3

    assert [step.key for step in levels[0]] == ['return_two.compute']
    assert [step.key for step in levels[1]] == ['add_three.compute', 'mult_three.compute']
    assert [step.key for step in levels[2]] == ['adder.compute']


def test_create_execution_plan_with_bad_inputs():
    with pytest.raises(DagsterInvalidConfigError):
        create_execution_plan(
            define_diamond_pipeline(), {'solids': {'add_three': {'inputs': {'num': 3}}}}
        )



import pytest

from dagster import (
    Any,
    Bool,
    Float,
    InputDefinition,
    Int,
    List,
    Optional,
    OutputDefinition,
    Path,
    DagsterInvalidConfigError,
    PipelineDefinition,
    String,
    execute_pipeline,
    lambda_solid,
)

from dagster.utils.test import get_temp_file_name


def _execute_pipeline_with_subset(pipeline, environment_dict, solid_subset):
    return execute_pipeline(
        pipeline.build_sub_pipeline(solid_subset), environment_dict=environment_dict
    )


def define_test_all_scalars_pipeline():
    @lambda_solid(input_defs=[InputDefinition('num', Int)])
    def take_int(num):
        return num

    @lambda_solid(output_def=OutputDefinition(Int))
    def produce_int():
        return 2

    @lambda_solid(input_defs=[InputDefinition('string', String)])
    def take_string(string):
        return string

    @lambda_solid(output_def=OutputDefinition(String))
    def produce_string():
        return 'foo'

    @lambda_solid(input_defs=[InputDefinition('path', Path)])
    def take_path(path):
        return path

    @lambda_solid(output_def=OutputDefinition(Path))
    def produce_path():
        return '/path/to/foo'

    @lambda_solid(input_defs=[InputDefinition('float_number', Float)])
    def take_float(float_number):
        return float_number

    @lambda_solid(output_def=OutputDefinition(Float))
    def produce_float():
        return 3.14

    @lambda_solid(input_defs=[InputDefinition('bool_value', Bool)])
    def take_bool(bool_value):
        return bool_value

    @lambda_solid(output_def=OutputDefinition(Bool))
    def produce_bool():
        return True

    @lambda_solid(input_defs=[InputDefinition('any_value', Any)])
    def take_any(any_value):
        return any_value

    @lambda_solid(output_def=OutputDefinition(Any))
    def produce_any():
        return True

    @lambda_solid(input_defs=[InputDefinition('string_list', List[String])])
    def take_string_list(string_list):
        return string_list

    @lambda_solid(input_defs=[InputDefinition('nullable_string', Optional[String])])
    def take_nullable_string(nullable_string):
        return nullable_string

    return PipelineDefinition(
        name='test_all_scalars_pipeline',
        solid_defs=[
            produce_any,
            produce_bool,
            produce_float,
            produce_int,
            produce_path,
            produce_string,
            take_any,
            take_bool,
            take_float,
            take_int,
            take_nullable_string,
            take_path,
            take_string,
            take_string_list,
        ],
    )


def single_input_env(solid_name, input_name, input_spec):
    return {'solids': {solid_name: {'inputs': {input_name: input_spec}}}}


def test_int_input_schema_value():
    result = _execute_pipeline_with_subset(
        define_test_all_scalars_pipeline(),
        environment_dict=single_input_env('take_int', 'num', {'value': 2}),
        solid_subset=['take_int'],
    )

    assert result.success
    assert result.result_for_solid('take_int').result_value() == 2


def test_int_input_schema_failure():
    with pytest.raises(DagsterInvalidConfigError) as exc_info:
        _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            environment_dict=single_input_env('take_int', 'num', {'value': 'dkjdfkdj'}),
            solid_subset=['take_int'],
        )

    assert 'Type failure at path "root:solids:take_int:inputs:num:value" on type "Int"' in str(
        exc_info.value
    )


def single_output_env(solid_name, output_spec):
    return {'solids': {solid_name: {'outputs': [{'result': output_spec}]}}}


def test_int_json_schema_roundtrip():
    with get_temp_file_name() as tmp_file:
        mat_result = _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            environment_dict=single_output_env('produce_int', {'json': {'path': tmp_file}}),
            solid_subset=['produce_int'],
        )

        assert mat_result.success

        source_result = _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            environment_dict=single_input_env('take_int', 'num', {'json': {'path': tmp_file}}),
            solid_subset=['take_int'],
        )

        assert source_result.result_for_solid('take_int').result_value() == 2


def test_int_pickle_schema_roundtrip():
    with get_temp_file_name() as tmp_file:
        mat_result = _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            environment_dict=single_output_env('produce_int', {'pickle': {'path': tmp_file}}),
            solid_subset=['produce_int'],
        )

        assert mat_result.success

        source_result = _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            environment_dict=single_input_env('take_int', 'num', {'pickle': {'path': tmp_file}}),
            solid_subset=['take_int'],
        )

        assert source_result.result_for_solid('take_int').result_value() == 2


def test_string_input_schema_value():
    result = _execute_pipeline_with_subset(
        define_test_all_scalars_pipeline(),
        environment_dict=single_input_env('take_string', 'string', {'value': 'dkjkfd'}),
        solid_subset=['take_string'],
    )

    assert result.success
    assert result.result_for_solid('take_string').result_value() == 'dkjkfd'


def test_string_input_schema_failure():
    with pytest.raises(DagsterInvalidConfigError) as exc_info:
        _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            environment_dict=single_input_env('take_string', 'string', {'value': 3343}),
            solid_subset=['take_string'],
        )

    assert (
        'Type failure at path "root:solids:take_string:inputs:string:value" on type "String"'
        in str(exc_info.value)
    )


def test_float_input_schema_value():
    result = _execute_pipeline_with_subset(
        define_test_all_scalars_pipeline(),
        environment_dict=single_input_env('take_float', 'float_number', {'value': 3.34}),
        solid_subset=['take_float'],
    )

    assert result.success
    assert result.result_for_solid('take_float').result_value() == 3.34


def test_float_input_schema_failure():
    with pytest.raises(DagsterInvalidConfigError) as exc_info:
        _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            environment_dict=single_input_env('take_float', 'float_number', {'value': '3343'}),
            solid_subset=['take_float'],
        )

    assert (
        'Type failure at path "root:solids:take_float:inputs:float_number:value" on type "Float"'
        in str(exc_info.value)
    )


def test_bool_input_schema_value():
    result = _execute_pipeline_with_subset(
        define_test_all_scalars_pipeline(),
        environment_dict=single_input_env('take_bool', 'bool_value', {'value': True}),
        solid_subset=['take_bool'],
    )

    assert result.success
    assert result.result_for_solid('take_bool').result_value() is True


def test_bool_input_schema_failure():
    with pytest.raises(DagsterInvalidConfigError) as exc_info:
        _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            environment_dict=single_input_env('take_bool', 'bool_value', {'value': '3343'}),
            solid_subset=['take_bool'],
        )

    assert (
        'Type failure at path "root:solids:take_bool:inputs:bool_value:value" on type "Bool".'
        in str(exc_info.value)
    )


def test_any_input_schema_value():
    result = _execute_pipeline_with_subset(
        define_test_all_scalars_pipeline(),
        environment_dict=single_input_env('take_any', 'any_value', {'value': 'ff'}),
        solid_subset=['take_any'],
    )

    assert result.success
    assert result.result_for_solid('take_any').result_value() == 'ff'

    result = _execute_pipeline_with_subset(
        define_test_all_scalars_pipeline(),
        environment_dict=single_input_env('take_any', 'any_value', {'value': 3843}),
        solid_subset=['take_any'],
    )

    assert result.success
    assert result.result_for_solid('take_any').result_value() == 3843


def test_none_string_input_schema_failure():
    with pytest.raises(DagsterInvalidConfigError) as exc_info:
        _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            environment_dict=single_input_env('take_string', 'string', None),
            solid_subset=['take_string'],
        )

    assert len(exc_info.value.errors) == 1

    error = exc_info.value.errors[0]

    print(error.message)

    assert error.message == (
        '''Must specify a field at path root:solids:take_string:inputs:string '''
        '''if more than one field is defined. Defined fields: ['json', 'pickle', 'value']'''
    )


def test_value_none_string_input_schema_failure():
    with pytest.raises(DagsterInvalidConfigError) as exc_info:
        _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            environment_dict=single_input_env('take_string', 'string', {'value': None}),
            solid_subset=['take_string'],
        )

    assert (
        'Type failure at path "root:solids:take_string:inputs:string:value" on type "String"'
        in str(exc_info.value)
    )


def test_string_json_schema_roundtrip():
    with get_temp_file_name() as tmp_file:
        mat_result = _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            environment_dict=single_output_env('produce_string', {'json': {'path': tmp_file}}),
            solid_subset=['produce_string'],
        )

        assert mat_result.success

        source_result = _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            environment_dict=single_input_env(
                'take_string', 'string', {'json': {'path': tmp_file}}
            ),
            solid_subset=['take_string'],
        )

        assert source_result.result_for_solid('take_string').result_value() == 'foo'


def test_string_pickle_schema_roundtrip():
    with get_temp_file_name() as tmp_file:
        mat_result = _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            environment_dict=single_output_env('produce_string', {'pickle': {'path': tmp_file}}),
            solid_subset=['produce_string'],
        )

        assert mat_result.success

        source_result = _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            environment_dict=single_input_env(
                'take_string', 'string', {'pickle': {'path': tmp_file}}
            ),
            solid_subset=['take_string'],
        )

        assert source_result.result_for_solid('take_string').result_value() == 'foo'


def test_path_input_schema_value():
    result = _execute_pipeline_with_subset(
        define_test_all_scalars_pipeline(),
        environment_dict=single_input_env('take_path', 'path', '/a/path'),
        solid_subset=['take_path'],
    )

    assert result.success
    assert result.result_for_solid('take_path').result_value() == '/a/path'


def test_path_input_schema_failure():
    with pytest.raises(DagsterInvalidConfigError) as exc_info:
        _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            environment_dict=single_input_env('take_path', 'path', {'value': 3343}),
            solid_subset=['take_path'],
        )

    assert 'Type failure at path "root:solids:take_path:inputs:path" on type "Path"' in str(
        exc_info.value
    )


def test_string_list_input():
    result = _execute_pipeline_with_subset(
        define_test_all_scalars_pipeline(),
        environment_dict=single_input_env('take_string_list', 'string_list', [{'value': 'foobar'}]),
        solid_subset=['take_string_list'],
    )

    assert result.success

    assert result.result_for_solid('take_string_list').result_value() == ['foobar']


def test_nullable_string_input_with_value():
    result = _execute_pipeline_with_subset(
        define_test_all_scalars_pipeline(),
        environment_dict=single_input_env(
            'take_nullable_string', 'nullable_string', {'value': 'foobar'}
        ),
        solid_subset=['take_nullable_string'],
    )

    assert result.success

    assert result.result_for_solid('take_nullable_string').result_value() == 'foobar'


def test_nullable_string_input_with_none_value():
    # Perhaps a confusing test case. Optional makes the entire enclosing structure nullable,
    # rather than the "value" value embedded within it
    with pytest.raises(DagsterInvalidConfigError) as exc_info:
        _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            environment_dict=single_input_env(
                'take_nullable_string', 'nullable_string', {'value': None}
            ),
            solid_subset=['take_nullable_string'],
        )

    assert (
        'Type failure at path "root:solids:take_nullable_string:inputs:nullable_string:value" '
        'on type "String"'
    ) in str(exc_info.value)


def test_nullable_string_input_without_value():
    result = _execute_pipeline_with_subset(
        define_test_all_scalars_pipeline(),
        environment_dict=single_input_env('take_nullable_string', 'nullable_string', None),
        solid_subset=['take_nullable_string'],
    )

    assert result.success

    assert result.result_for_solid('take_nullable_string').result_value() is None



from dagster import PipelineDefinition, OutputDefinition, lambda_solid, String
from dagster.core.definitions import create_environment_schema


def test_materialization_schema_types():
    @lambda_solid(output_def=OutputDefinition(String))
    def return_one():
        return 1

    pipeline_def = PipelineDefinition(
        name='test_materialization_schema_types', solid_defs=[return_one]
    )

    environment_schema = create_environment_schema(pipeline_def)

    string_mat_schema = environment_schema.config_type_named('String.MaterializationSchema')

    string_json_mat_schema = string_mat_schema.fields['json'].config_type

    assert environment_schema.config_type_keyed(string_json_mat_schema.key)



from __future__ import print_function

import os

import pytest

from click.testing import CliRunner

from dagster import lambda_solid, RepositoryDefinition, pipeline, DagsterInvariantViolationError
from dagster.core.storage.runs import base_runs_directory
from dagster.cli.load_handle import CliUsageError
from dagster.cli.pipeline import (
    execute_print_command,
    execute_list_command,
    execute_execute_command,
    execute_scaffold_command,
    pipeline_execute_command,
    pipeline_list_command,
    pipeline_print_command,
    pipeline_scaffold_command,
)
from dagster.utils import script_relative_path


def no_print(_):
    return None


@lambda_solid
def do_something():
    return 1


@pipeline(name='foo')
def foo_pipeline():
    do_something()


def define_foo_pipeline():
    return foo_pipeline


def define_bar_repo():
    return RepositoryDefinition('bar', {'foo': define_foo_pipeline})


def test_list_command():
    runner = CliRunner()

    execute_list_command(
        {
            'repository_yaml': None,
            'python_file': script_relative_path('test_cli_commands.py'),
            'module_name': None,
            'fn_name': 'define_bar_repo',
        },
        no_print,
    )

    result = runner.invoke(
        pipeline_list_command,
        ['-f', script_relative_path('test_cli_commands.py'), '-n', 'define_bar_repo'],
    )
    assert result.exit_code == 0
    assert result.output == (
        'Repository bar\n'
        '**************\n'
        'Pipeline: foo\n'
        'Solids: (Execution Order)\n'
        '    do_something\n'
    )

    execute_list_command(
        {
            'repository_yaml': None,
            'python_file': None,
            'module_name': 'dagster_examples.intro_tutorial.repos',
            'fn_name': 'define_repo',
        },
        no_print,
    )

    result = runner.invoke(
        pipeline_list_command, ['-m', 'dagster_examples.intro_tutorial.repos', '-n', 'define_repo']
    )
    assert result.exit_code == 0
    assert result.output == (
        'Repository demo_repository\n'
        '**************************\n'
        'Pipeline: repo_demo_pipeline\n'
        'Solids: (Execution Order)\n'
        '    hello_world\n'
    )

    execute_list_command(
        {
            'repository_yaml': script_relative_path('repository.yaml'),
            'python_file': None,
            'module_name': None,
            'fn_name': None,
        },
        no_print,
    )

    result = runner.invoke(pipeline_list_command, ['-y', script_relative_path('repository.yaml')])
    assert result.exit_code == 0
    assert result.output == (
        'Repository demo_repository\n'
        '**************************\n'
        'Pipeline: repo_demo_pipeline\n'
        'Solids: (Execution Order)\n'
        '    hello_world\n'
    )

    with pytest.raises(CliUsageError):
        execute_list_command(
            {
                'repository_yaml': None,
                'python_file': 'foo.py',
                'module_name': 'dagster_examples.intro_tutorial.repos',
                'fn_name': 'define_repo',
            },
            no_print,
        )

    result = runner.invoke(
        pipeline_list_command,
        ['-f', 'foo.py', '-m', 'dagster_examples.intro_tutorial.repos', '-n', 'define_repo'],
    )
    assert result.exit_code == 1
    assert isinstance(result.exception, CliUsageError)

    with pytest.raises(CliUsageError):
        execute_list_command(
            {
                'repository_yaml': None,
                'python_file': None,
                'module_name': 'dagster_examples.intro_tutorial.repos',
                'fn_name': None,
            },
            no_print,
        )

    result = runner.invoke(pipeline_list_command, ['-m', 'dagster_examples.intro_tutorial.repos'])
    assert result.exit_code == 1
    assert isinstance(result.exception, CliUsageError)

    with pytest.raises(CliUsageError):
        execute_list_command(
            {
                'repository_yaml': None,
                'python_file': script_relative_path('test_cli_commands.py'),
                'module_name': None,
                'fn_name': None,
            },
            no_print,
        )

    result = runner.invoke(
        pipeline_list_command, ['-f', script_relative_path('test_cli_commands.py')]
    )
    assert result.exit_code == 1
    assert isinstance(result.exception, CliUsageError)


def valid_execute_args():
    return [
        {
            'repository_yaml': script_relative_path('repository_file.yaml'),
            'pipeline_name': ('foo',),
            'python_file': None,
            'module_name': None,
            'fn_name': None,
        },
        {
            'repository_yaml': script_relative_path('repository_module.yaml'),
            'pipeline_name': ('repo_demo_pipeline',),
            'python_file': None,
            'module_name': None,
            'fn_name': None,
        },
        {
            'repository_yaml': None,
            'pipeline_name': ('foo',),
            'python_file': script_relative_path('test_cli_commands.py'),
            'module_name': None,
            'fn_name': 'define_bar_repo',
        },
        {
            'repository_yaml': None,
            'pipeline_name': ('repo_demo_pipeline',),
            'python_file': None,
            'module_name': 'dagster_examples.intro_tutorial.repos',
            'fn_name': 'define_repo',
        },
        {
            'repository_yaml': None,
            'pipeline_name': (),
            'python_file': None,
            'module_name': 'dagster_examples.intro_tutorial.repos',
            'fn_name': 'repo_demo_pipeline',
        },
        {
            'repository_yaml': None,
            'pipeline_name': (),
            'python_file': script_relative_path('test_cli_commands.py'),
            'module_name': None,
            'fn_name': 'define_foo_pipeline',
        },
        {
            'repository_yaml': None,
            'pipeline_name': (),
            'python_file': script_relative_path('test_cli_commands.py'),
            'module_name': None,
            'fn_name': 'foo_pipeline',
        },
    ]


def valid_cli_args():
    return [
        ['-y', script_relative_path('repository_file.yaml'), 'foo'],
        ['-y', script_relative_path('repository_module.yaml'), 'repo_demo_pipeline'],
        ['-f', script_relative_path('test_cli_commands.py'), '-n', 'define_bar_repo', 'foo'],
        ['-m', 'dagster_examples.intro_tutorial.repos', '-n', 'define_repo', 'repo_demo_pipeline'],
        ['-m', 'dagster_examples.intro_tutorial.repos', '-n', 'repo_demo_pipeline'],
        ['-f', script_relative_path('test_cli_commands.py'), '-n', 'define_foo_pipeline'],
    ]


def test_print_command():
    for cli_args in valid_execute_args():
        execute_print_command(verbose=True, cli_args=cli_args, print_fn=no_print)

    for cli_args in valid_execute_args():
        execute_print_command(verbose=False, cli_args=cli_args, print_fn=no_print)

    runner = CliRunner()

    for cli_args in valid_cli_args():
        result = runner.invoke(pipeline_print_command, cli_args)
        assert result.exit_code == 0

        result = runner.invoke(pipeline_print_command, ['--verbose'] + cli_args)
        assert result.exit_code == 0


def test_execute_mode_command():
    runner = CliRunner()

    add_result = runner_pipeline_execute(
        runner,
        [
            '-y',
            script_relative_path('../repository.yaml'),
            '--env',
            script_relative_path('../environments/multi_mode_with_resources/add_mode.yaml'),
            '-d',
            'add_mode',
            'multi_mode_with_resources',  # pipeline name
        ],
    )

    # this is quite horrific and fragile easiest way to check for correcness
    assert 'value 5' in add_result.stdout

    mult_result = runner_pipeline_execute(
        runner,
        [
            '-y',
            script_relative_path('../repository.yaml'),
            '--env',
            script_relative_path('../environments/multi_mode_with_resources/mult_mode.yaml'),
            '-d',
            'mult_mode',
            'multi_mode_with_resources',  # pipeline name
        ],
    )

    assert 'value 12' in mult_result.stdout

    double_adder_result = runner_pipeline_execute(
        runner,
        [
            '-y',
            script_relative_path('../repository.yaml'),
            '--env',
            script_relative_path(
                '../environments/multi_mode_with_resources/double_adder_mode.yaml'
            ),
            '-d',
            'double_adder_mode',
            'multi_mode_with_resources',  # pipeline name
        ],
    )

    assert 'value 25' in double_adder_result.stdout


def test_execute_preset_command():
    runner = CliRunner()
    add_result = runner_pipeline_execute(
        runner,
        [
            '-y',
            script_relative_path('../repository.yaml'),
            '-p',
            'add',
            'multi_mode_with_resources',  # pipeline name
        ],
    )
    # this is quite horrific and fragile easiest way to check for correcness
    assert 'value 5' in add_result.stdout


def test_execute_command():
    for cli_args in valid_execute_args():
        execute_execute_command(env=None, raise_on_error=True, cli_args=cli_args)

    for cli_args in valid_execute_args():
        execute_execute_command(
            env=[script_relative_path('default_log_error_env.yaml')],
            raise_on_error=True,
            cli_args=cli_args,
        )

    runner = CliRunner()

    for cli_args in valid_cli_args():
        runner_pipeline_execute(runner, cli_args)

        runner_pipeline_execute(
            runner, ['--env', script_relative_path('default_log_error_env.yaml')] + cli_args
        )


def test_fn_not_found_execute():
    with pytest.raises(DagsterInvariantViolationError) as exc_info:
        execute_execute_command(
            env=None,
            raise_on_error=True,
            cli_args={
                'repository_yaml': None,
                'pipeline_name': (),
                'python_file': script_relative_path('test_cli_commands.py'),
                'module_name': None,
                'fn_name': 'nope',
            },
        )

    assert 'nope not found in module' in str(exc_info.value)


def not_a_repo_or_pipeline_fn():
    return 'kdjfkjdf'


not_a_repo_or_pipeline = 123


def test_fn_is_wrong_thing():
    with pytest.raises(DagsterInvariantViolationError) as exc_info:
        execute_execute_command(
            env=None,
            raise_on_error=True,
            cli_args={
                'repository_yaml': None,
                'pipeline_name': (),
                'python_file': script_relative_path('test_cli_commands.py'),
                'module_name': None,
                'fn_name': 'not_a_repo_or_pipeline',
            },
        )

    assert str(exc_info.value) == (
        'not_a_repo_or_pipeline must be a function that returns a '
        'PipelineDefinition or a RepositoryDefinition, or a function '
        'decorated with @pipeline.'
    )


def test_fn_returns_wrong_thing():
    with pytest.raises(DagsterInvariantViolationError) as exc_info:
        execute_execute_command(
            env=None,
            raise_on_error=True,
            cli_args={
                'repository_yaml': None,
                'pipeline_name': (),
                'python_file': script_relative_path('test_cli_commands.py'),
                'module_name': None,
                'fn_name': 'not_a_repo_or_pipeline_fn',
            },
        )

    assert str(exc_info.value) == (
        'not_a_repo_or_pipeline_fn is a function but must return a '
        'PipelineDefinition or a RepositoryDefinition, or be decorated '
        'with @pipeline.'
    )


def runner_pipeline_execute(runner, cli_args):
    result = runner.invoke(pipeline_execute_command, cli_args)
    if result.exit_code != 0:
        # CliRunner captures stdout so printing it out here
        raise Exception(
            (
                'dagster pipeline execute commands with cli_args {cli_args} '
                'returned exit_code {exit_code} with stdout:\n"{stdout}" and '
                '\nresult as string: "{result}"'
            ).format(
                cli_args=cli_args, exit_code=result.exit_code, stdout=result.stdout, result=result
            )
        )
    return result


def test_scaffold_command():
    for cli_args in valid_execute_args():
        cli_args['print_only_required'] = True
        execute_scaffold_command(cli_args=cli_args, print_fn=no_print)

        cli_args['print_only_required'] = False
        execute_scaffold_command(cli_args=cli_args, print_fn=no_print)

    runner = CliRunner()

    for cli_args in valid_cli_args():
        result = runner.invoke(pipeline_scaffold_command, cli_args)
        assert result.exit_code == 0

        result = runner.invoke(pipeline_scaffold_command, ['-p'] + cli_args)
        assert result.exit_code == 0


def test_default_memory_run_storage():
    cli_args = {
        'repository_yaml': script_relative_path('repository_file.yaml'),
        'pipeline_name': ('foo',),
        'python_file': None,
        'module_name': None,
        'fn_name': None,
    }
    result = execute_execute_command(env=None, raise_on_error=True, cli_args=cli_args)
    assert result.success

    run_dir = os.path.join(base_runs_directory(), result.run_id)

    assert not os.path.isdir(run_dir)


def test_override_with_in_memory_storage():
    cli_args = {
        'repository_yaml': script_relative_path('repository_file.yaml'),
        'pipeline_name': ('foo',),
        'python_file': None,
        'module_name': None,
        'fn_name': None,
    }
    result = execute_execute_command(
        env=[script_relative_path('in_memory_env.yaml')], raise_on_error=True, cli_args=cli_args
    )
    assert result.success

    run_dir = os.path.join(base_runs_directory(), result.run_id)

    assert not os.path.exists(run_dir)


def test_override_with_filesystem_storage():
    cli_args = {
        'repository_yaml': script_relative_path('repository_file.yaml'),
        'pipeline_name': ('foo',),
        'python_file': None,
        'module_name': None,
        'fn_name': None,
    }
    result = execute_execute_command(
        env=[script_relative_path('filesystem_env.yaml')], raise_on_error=True, cli_args=cli_args
    )
    assert result.success

    run_dir = os.path.join(base_runs_directory(), result.run_id)

    assert os.path.exists(run_dir)



import pytest

from dagster.cli.structured_flags import (
    structure_flags,
    SingleArgument,
    StructuredArguments,
    NamedArgument,
    NamedKeyArgument,
    StructuredArgumentsError,
)


def test_structure_flag_empty():
    assert structure_flags([]) is None


def test_structure_flag_single():
    foo_argument = structure_flags(['foo'])
    assert isinstance(foo_argument, StructuredArguments)
    assert isinstance(foo_argument.single_argument, SingleArgument)
    assert not foo_argument.named_arguments
    assert not foo_argument.named_key_arguments
    assert foo_argument.single_argument.value == 'foo'


def test_structure_double_single_fails():
    with pytest.raises(StructuredArgumentsError):
        structure_flags(['foo', 'bar'])

    with pytest.raises(StructuredArgumentsError):
        structure_flags(['foo', 'bar=ad'])

    with pytest.raises(StructuredArgumentsError):
        structure_flags(['foo', 'bar.baz=ad'])


def test_structure_flag_single_named():
    structured_arguments = structure_flags(['foo=bar'])
    assert isinstance(structured_arguments, StructuredArguments)
    assert len(structured_arguments.named_arguments) == 1
    assert not structured_arguments.named_key_arguments

    named_argument = structured_arguments.named_arguments[0]

    assert isinstance(named_argument, NamedArgument)
    assert named_argument.name == 'foo'
    assert named_argument.value == 'bar'


def test_structure_flag_single_named_equal_in_name():
    structured_arguments = structure_flags(['foo=bar='])
    assert isinstance(structured_arguments, StructuredArguments)
    assert len(structured_arguments.named_arguments) == 1
    assert not structured_arguments.named_key_arguments

    named_argument = structured_arguments.named_arguments[0]

    assert isinstance(named_argument, NamedArgument)
    assert named_argument.name == 'foo'
    assert named_argument.value == 'bar='


def test_structure_flag_single_named_key():
    structured_arguments = structure_flags(['foo.baz=bar'])
    assert isinstance(structured_arguments, StructuredArguments)
    assert len(structured_arguments.named_key_arguments) == 1
    assert not structured_arguments.named_arguments

    named_key_argument = structured_arguments.named_key_arguments[0]

    assert isinstance(named_key_argument, NamedKeyArgument)
    assert named_key_argument.name == 'foo'
    assert named_key_argument.key == 'baz'
    assert named_key_argument.value == 'bar'


def test_structure_flag_double_named():
    structured_arguments = structure_flags(['foo=foo_arg', 'bar=bar_arg'])
    assert not structured_arguments.single_argument
    assert len(structured_arguments.named_arguments) == 2
    assert not structured_arguments.named_key_arguments

    assert structured_arguments.named_arguments[0].name == 'foo'
    assert structured_arguments.named_arguments[0].value == 'foo_arg'

    assert structured_arguments.named_arguments[1].name == 'bar'
    assert structured_arguments.named_arguments[1].value == 'bar_arg'


def test_structure_flag_double_dup_fails():
    with pytest.raises(StructuredArgumentsError, match='Cannot specify name twice'):
        structure_flags(['foo=foo_arg', 'foo=bar_arg'])

    with pytest.raises(StructuredArgumentsError):
        structure_flags(['foo=foo_arg', 'foo.key=bar_arg'])

    with pytest.raises(StructuredArgumentsError):
        structure_flags(['foo.key=foo_arg', 'foo=bar_arg'])

    with pytest.raises(StructuredArgumentsError, match='Cannot specify key'):
        structure_flags(['foo.key=foo_arg', 'foo.key=bar_arg'])


def test_structure_two_keys():
    structured_arguments = structure_flags(['foo.key1=key1_arg', 'foo.key2=key2_arg'])
    assert not structured_arguments.single_argument
    assert not structured_arguments.named_arguments
    assert len(structured_arguments.named_key_arguments) == 2

    assert structured_arguments.named_key_arguments[0].name == 'foo'
    assert structured_arguments.named_key_arguments[0].key == 'key1'
    assert structured_arguments.named_key_arguments[0].value == 'key1_arg'

    assert structured_arguments.named_key_arguments[1].name == 'foo'
    assert structured_arguments.named_key_arguments[1].key == 'key2'
    assert structured_arguments.named_key_arguments[1].value == 'key2_arg'



import logging

from dagster import PipelineDefinition
from dagster.core.log_manager import DagsterLogManager
from dagster.core.execution.context.logger import InitLoggerContext
from dagster.utils.log import construct_single_handler_logger


class LogTestHandler(logging.Handler):
    def __init__(self, records):
        self.records = records
        super(LogTestHandler, self).__init__()

    def emit(self, record):
        self.records.append(record)


def test_log_level_filtering():
    records = []
    critical_records = []

    debug_logger_def = construct_single_handler_logger(
        'debug_handler', 'debug', LogTestHandler(records)
    )
    critical_logger_def = construct_single_handler_logger(
        'critical_handler', 'critical', LogTestHandler(critical_records)
    )

    loggers = [
        logger_def.logger_fn(InitLoggerContext({}, PipelineDefinition([]), logger_def, ''))
        for logger_def in [debug_logger_def, critical_logger_def]
    ]

    log_manager = DagsterLogManager('', {}, loggers)

    log_manager.debug('Hello, there!')

    messages = [x.dagster_meta['orig_message'] for x in records]

    assert 'Hello, there!' in messages

    critical_messages = [x.dagster_meta['orig_message'] for x in critical_records]

    assert 'Hello, there!' not in critical_messages



from dagster import PipelineDefinition, SolidDefinition, Int, Field, String, List, Optional, Dict

from dagster.core.types.field import resolve_to_config_type
from dagster.core.types.type_printer import print_type_to_string


def assert_inner_types(parent_type, *dagster_types):
    assert set(list(map(lambda t: t.name, resolve_to_config_type(parent_type).inner_types))) == set(
        map(lambda x: x.name, map(resolve_to_config_type, dagster_types))
    )


def test_basic_type_print():
    assert print_type_to_string(Int) == 'Int'
    assert_inner_types(Int)


def test_basic_list_type_print():
    assert print_type_to_string(List[Int]) == '[Int]'
    assert_inner_types(List[Int], Int)


def test_double_list_type_print():
    assert print_type_to_string(List[List[Int]]) == '[[Int]]'
    int_list = List[Int]
    list_int_list = List[int_list]
    assert_inner_types(list_int_list, Int, int_list)


def test_basic_nullable_type_print():
    assert print_type_to_string(Optional[Int]) == 'Int?'
    nullable_int = Optional[Int]
    assert_inner_types(nullable_int, Int)


def test_nullable_list_combos():
    assert print_type_to_string(List[Int]) == '[Int]'
    assert print_type_to_string(Optional[List[Int]]) == '[Int]?'
    assert print_type_to_string(List[Optional[Int]]) == '[Int?]'
    assert print_type_to_string(Optional[List[Optional[Int]]]) == '[Int?]?'


def test_basic_dict():
    output = print_type_to_string(Dict({'int_field': Field(Int)}))

    expected = '''{
  int_field: Int
}'''

    assert output == expected


def test_two_field_dicts():
    two_field_dict = Dict({'int_field': Field(Int), 'string_field': Field(String)})
    assert_inner_types(two_field_dict, Int, String)

    output = print_type_to_string(two_field_dict)

    expected = '''{
  int_field: Int
  string_field: String
}'''

    assert output == expected


def test_two_field_dicts_same_type():
    two_field_dict = Dict({'int_field1': Field(Int), 'int_field2': Field(Int)})
    assert_inner_types(two_field_dict, Int)

    output = print_type_to_string(two_field_dict)

    expected = '''{
  int_field1: Int
  int_field2: Int
}'''

    assert output == expected


def test_optional_field():
    output = print_type_to_string(Dict({'int_field': Field(Int, is_optional=True)}))

    expected = '''{
  int_field?: Int
}'''

    assert output == expected


def test_single_level_dict_lists_and_nullable():
    output = print_type_to_string(
        Dict(
            {
                'nullable_int_field': Field(Optional[Int]),
                'optional_int_field': Field(Int, is_optional=True),
                'string_list_field': Field(List[String]),
            }
        )
    )

    expected = '''{
  nullable_int_field: Int?
  optional_int_field?: Int
  string_list_field: [String]
}'''

    assert output == expected


def test_nested_dict():
    nested_type = Dict({'int_field': Field(Int)})
    outer_type = Dict({'nested': Field(nested_type)})
    output = print_type_to_string(outer_type)

    assert_inner_types(outer_type, Int, nested_type)

    expected = '''{
  nested: {
    int_field: Int
  }
}'''

    assert output == expected


def test_test_type_pipeline_construction():
    assert define_test_type_pipeline()


def define_solid_for_test_type(name, runtime_type):
    return SolidDefinition(
        name=name,
        input_defs=[],
        output_defs=[],
        config_field=Field(runtime_type),
        compute_fn=lambda _info, _inputs: None,
    )


# launch in dagit with this command:
# dagit -f test_type_printer.py -n define_test_type_pipeline
def define_test_type_pipeline():
    return PipelineDefinition(
        name='test_type_pipeline',
        solid_defs=[
            define_solid_for_test_type('int_config', Int),
            define_solid_for_test_type('list_of_int_config', List[Int]),
            define_solid_for_test_type('nullable_list_of_int_config', Optional[List[Int]]),
            define_solid_for_test_type('list_of_nullable_int_config', List[Optional[Int]]),
            define_solid_for_test_type(
                'nullable_list_of_nullable_int_config', Optional[List[Optional[Int]]]
            ),
            define_solid_for_test_type(
                'simple_dict', Dict({'int_field': Field(Int), 'string_field': Field(String)})
            ),
            define_solid_for_test_type(
                'dict_with_optional_field',
                Dict(
                    {
                        'nullable_int_field': Field(Optional[Int]),
                        'optional_int_field': Field(Int, is_optional=True),
                        'string_list_field': Field(List[String]),
                    }
                ),
            ),
            define_solid_for_test_type(
                'nested_dict', Dict({'nested': Field(Dict({'int_field': Field(Int)}))})
            ),
        ],
    )



import datetime
import uuid
import toposort as toposort_

AIRFLOW_TS_DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%f+00:00"


def convert_airflow_datestr_to_epoch_ts(airflow_ts):
    '''convert_airflow_datestr_to_epoch_ts
    Converts Airflow time strings (e.g. 2019-06-26T17:19:09.989485) to epoch timestamps.

    See: https://stackoverflow.com/a/50650878/11295366
    dt.timestamp() is not available in py2
    '''
    dt = datetime.datetime.strptime(airflow_ts, AIRFLOW_TS_DATETIME_FORMAT)
    return (dt - datetime.datetime(1970, 1, 1)).total_seconds()


def toposort(data):
    return [sorted(list(level)) for level in toposort_.toposort(data)]


def toposort_flatten(data):
    return [item for level in toposort(data) for item in level]


def make_new_run_id():
    return str(uuid.uuid4())



from contextlib import contextmanager

from dagster import check, seven


class TimerResult(object):
    def __init__(self):
        self.start_time = seven.time_fn()
        self.end_time = None

    @property
    def seconds(self):
        check.invariant(self.end_time is not None, 'end time is not set')
        return self.end_time - self.start_time

    @property
    def millis(self):
        return self.seconds * 1000


@contextmanager
def time_execution_scope():
    '''Usage:

    from solid_util.timing import time_execution_scope
    with time_execution_scope() as timer_result:
        do_some_operation()

    print(
        'do_some_operation took {timer_result.millis} milliseconds'.format(
            timer_result=timer_result
        )
    )
    '''

    timer_result = TimerResult()
    yield timer_result
    timer_result.end_time = seven.time_fn()



from collections import namedtuple

from dagster import check

from dagster.core.log_manager import DagsterLogManager
from dagster.core.definitions.pipeline import PipelineDefinition
from dagster.core.definitions.resource import ResourceDefinition


class InitResourceContext(
    namedtuple(
        'InitResourceContext', 'resource_config pipeline_def resource_def run_id log_manager'
    )
):
    '''
    Resource-specific initialization context.

    Attributes:
        resource_config (Any):
            The configuration data provided by the environment config. The schema for this
            data is defined by ``config_field`` on the :py:class:`ResourceDefinition`
        pipeline_def (PipelineDefinition): The pipeline definition currently being executed.
        resource_def (ResourceDefinition): The resource definition for the resource being
            constructed.
        run_id (str): The ID for this run of the pipeline.
        log_manager (DagsterLogManager): The log manager for this run of the pipeline
    '''

    def __new__(cls, resource_config, pipeline_def, resource_def, run_id, log_manager=None):
        return super(InitResourceContext, cls).__new__(
            cls,
            resource_config,
            check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition),
            check.inst_param(resource_def, 'resource_def', ResourceDefinition),
            check.str_param(run_id, 'run_id'),
            check.opt_inst_param(log_manager, 'log_manager', DagsterLogManager),
        )



from dagster.core.utils import convert_airflow_datestr_to_epoch_ts

EPS = 1e-6


def test_airflow_ts():
    example = '2019-06-26T17:19:09.989485+00:00'
    res = convert_airflow_datestr_to_epoch_ts(example)
    assert (res - 1561569549.989485) < EPS



from contextlib import contextmanager

import os
import tempfile
import time
import shutil

from dagster import PipelineDefinition, solid, execute_pipeline
from dagster.core.storage.runs import (
    DagsterRunMeta,
    FileSystemRunStorage,
    InMemoryRunStorage,
    base_runs_directory,
)


@contextmanager
def temp_run_storage():
    try:
        base_dir = tempfile.mkdtemp()
        yield FileSystemRunStorage(base_dir)
    finally:
        if os.path.exists(base_dir):
            shutil.rmtree(base_dir)


def test_default_run_storage():
    @solid
    def check_run_storage(context):
        assert isinstance(context.get_system_context().run_storage, InMemoryRunStorage)

    pipeline = PipelineDefinition(name='default_run_storage_test', solid_defs=[check_run_storage])

    result = execute_pipeline(pipeline)

    assert result.success

    assert not os.path.exists(os.path.join(base_runs_directory(), result.run_id))


def test_config_specified_filesystem_run_storage():
    @solid
    def check_run_storage(context):
        assert isinstance(context.get_system_context().run_storage, FileSystemRunStorage)

    pipeline = PipelineDefinition(name='default_run_storage_test', solid_defs=[check_run_storage])

    result = execute_pipeline(pipeline, environment_dict={'storage': {'filesystem': {}}})

    assert result.success

    assert os.path.exists(os.path.join(base_runs_directory(), result.run_id))


def test_empty_storage():
    with temp_run_storage() as run_storage:
        assert run_storage.get_run_ids() == []


def test_filesystem_persist_one_run():
    with temp_run_storage() as run_storage:
        do_test_single_write_read(run_storage)


def test_in_memory_persist_one_run():
    do_test_single_write_read(InMemoryRunStorage())


def do_test_single_write_read(run_storage):
    run_id = 'some_run_id'
    current_time = time.time()
    run_storage.write_dagster_run_meta(
        DagsterRunMeta(run_id=run_id, timestamp=current_time, pipeline_name='some_pipeline')
    )

    run_meta = run_storage.get_run_meta(run_id)

    assert run_meta.run_id == run_id
    assert run_meta.timestamp == current_time
    assert run_meta.pipeline_name == 'some_pipeline'

    assert run_storage.get_run_metas() == [run_meta]

    run_storage.nuke()

    assert run_storage.get_run_metas() == []



from collections import namedtuple
from dagster.core.types.decorator import (
    dagster_type,
    get_runtime_type_on_decorated_klass,
    as_dagster_type,
)


def test_dagster_type_decorator():
    @dagster_type(name=None)
    class Foo(object):
        pass

    @dagster_type()
    class Bar(object):
        pass

    @dagster_type
    class Baaz(object):
        pass

    assert get_runtime_type_on_decorated_klass(Foo).name == 'Foo'
    assert get_runtime_type_on_decorated_klass(Bar).name == 'Bar'
    assert get_runtime_type_on_decorated_klass(Baaz).name == 'Baaz'


def test_dagster_type_decorator_name_desc():
    @dagster_type(name='DifferentName', description='desc')
    class Something(object):
        pass

    runtime_type = get_runtime_type_on_decorated_klass(Something)
    assert runtime_type.name == 'DifferentName'
    assert runtime_type.description == 'desc'


def test_make_dagster_type():
    OverwriteNameTuple = as_dagster_type(namedtuple('SomeNamedTuple', 'prop'))
    runtime_type = get_runtime_type_on_decorated_klass(OverwriteNameTuple)
    assert runtime_type.name == 'SomeNamedTuple'
    assert OverwriteNameTuple(prop='foo').prop == 'foo'

    OverwriteNameTuple = as_dagster_type(namedtuple('SomeNamedTuple', 'prop'), name='OverwriteName')
    runtime_type = get_runtime_type_on_decorated_klass(OverwriteNameTuple)
    assert runtime_type.name == 'OverwriteName'
    assert OverwriteNameTuple(prop='foo').prop == 'foo'



from dagster import check
from dagster.core.errors import DagsterInvalidDefinitionError
from .builtin_enum import BuiltinEnum
from .config import Any, ConfigType, List, Nullable
from .field_utils import FieldImpl, FIELD_NO_DEFAULT_PROVIDED, INFER_OPTIONAL_COMPOSITE_FIELD
from .wrapping import WrappingListType, WrappingNullableType


def resolve_to_config_list(list_type):
    check.inst_param(list_type, 'list_type', WrappingListType)
    return List(resolve_to_config_type(list_type.inner_type))


def resolve_to_config_nullable(nullable_type):
    check.inst_param(nullable_type, 'nullable_type', WrappingNullableType)
    return Nullable(resolve_to_config_type(nullable_type.inner_type))


def resolve_to_config_type(dagster_type):
    from .mapping import remap_python_type

    dagster_type = remap_python_type(dagster_type)

    if dagster_type is None:
        return Any.inst()
    if BuiltinEnum.contains(dagster_type):
        return ConfigType.from_builtin_enum(dagster_type)
    if isinstance(dagster_type, WrappingListType):
        return resolve_to_config_list(dagster_type).inst()
    if isinstance(dagster_type, WrappingNullableType):
        return resolve_to_config_nullable(dagster_type).inst()
    if isinstance(dagster_type, type) and issubclass(dagster_type, ConfigType):
        return dagster_type.inst()

    return None


def Field(
    dagster_type,
    default_value=FIELD_NO_DEFAULT_PROVIDED,
    is_optional=INFER_OPTIONAL_COMPOSITE_FIELD,
    is_secret=False,
    description=None,
):
    '''
    The schema for configuration data that describes the type, optionality, defaults, and description.

    Args:
        dagster_type (DagsterType):
            A ``DagsterType`` describing the schema of this field, ie `Dict({'example': Field(String)})`
        default_value (Any):
            A default value to use that respects the schema provided via dagster_type
        is_optional (bool): Whether the presence of this field is optional
        description (str):
    '''
    config_type = resolve_to_config_type(dagster_type)
    if not config_type:
        raise DagsterInvalidDefinitionError(
            (
                'Attempted to pass {value_repr} to a Field that expects a valid '
                'dagster type usable in config (e.g. Dict, NamedDict, Int, String et al).'
            ).format(value_repr=repr(dagster_type))
        )
    return FieldImpl(
        config_type=config_type,
        default_value=default_value,
        is_optional=is_optional,
        is_secret=is_secret,
        description=description,
    )



from collections import namedtuple

from dagster import check
from dagster.core.definitions import (
    CompositeSolidDefinition,
    InputDefinition,
    OutputDefinition,
    PipelineDefinition,
    ModeDefinition,
    Solid,
    SolidDefinition,
    SolidHandle,
    SolidOutputHandle,
    solids_in_topological_order,
)
from dagster.core.definitions.dependency import DependencyStructure
from dagster.core.errors import DagsterInvariantViolationError
from dagster.core.system_config.objects import EnvironmentConfig
from dagster.core.utils import toposort

from .compute import create_compute_step
from .expectations import create_expectations_subplan, decorate_with_expectations
from .objects import ExecutionStep, StepInput, StepKind, StepOutputHandle
from .utility import create_join_outputs_step


class _PlanBuilder:
    '''_PlanBuilder. This is the state that is built up during the execution plan build process.

    steps List[ExecutionStep]: a list of the execution steps that have been created.

    step_output_map Dict[SolidOutputHandle, StepOutputHandle]:  maps logical solid outputs
    (solid_name, output_name) to particular step outputs. This covers the case where a solid maps to
    multiple steps and one wants to be able to attach to the logical output of a solid during
    execution.
    '''

    def __init__(self, pipeline_def, environment_config, mode_definition):
        self.pipeline_def = check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
        self.environment_config = check.inst_param(
            environment_config, 'environment_config', EnvironmentConfig
        )
        self.mode_definition = check.inst_param(mode_definition, 'mode_definition', ModeDefinition)
        self.steps = []
        self.step_output_map = dict()
        self.seen_keys = set()

    @property
    def pipeline_name(self):
        return self.pipeline_def.name

    def add_step(self, step):
        # Keep track of the step keys we've seen so far to ensure we don't add duplicates
        if step.key in self.seen_keys:
            keys = [s.key for s in self.steps]
            check.failed(
                'Duplicated key {key}. Full list seen so far: {key_list}.'.format(
                    key=step.key, key_list=keys
                )
            )
        self.seen_keys.add(step.key)
        self.steps.append(step)

    def add_steps(self, steps):
        for step in steps:
            self.add_step(step)

    def get_output_handle(self, key):
        check.inst_param(key, 'key', SolidOutputHandle)
        return self.step_output_map[key]

    def set_output_handle(self, key, val):
        check.inst_param(key, 'key', SolidOutputHandle)
        check.inst_param(val, 'val', StepOutputHandle)
        self.step_output_map[key] = val

    def build(self):
        '''Builds the execution plan.
        '''

        # Recursively build the exeuction plan starting at the root pipeline
        self._build_from_sorted_solids(
            solids_in_topological_order(self.pipeline_def), self.pipeline_def.dependency_structure
        )

        # Construct dependency dictionary
        deps = {step.key: set() for step in self.steps}

        for step in self.steps:
            for step_input in step.step_inputs:
                if step_input.is_from_output:
                    deps[step.key].add(step_input.prev_output_handle.step_key)

        step_dict = {step.key: step for step in self.steps}

        system_storage_def = self.mode_definition.get_system_storage_def(
            self.environment_config.storage.system_storage_name
        )

        return ExecutionPlan(self.pipeline_def, step_dict, deps, system_storage_def.is_persistent)

    def _build_from_sorted_solids(
        self, solids, dependency_structure, parent_handle=None, parent_step_inputs=None
    ):
        terminal_transform_step = None
        for solid in solids:
            handle = SolidHandle(solid.name, solid.definition.name, parent_handle)

            ### 1. INPUTS
            # Create and add execution plan steps for solid inputs
            step_inputs = []
            for input_name, input_def in solid.definition.input_dict.items():
                step_input = get_step_input(
                    self,
                    solid,
                    input_name,
                    input_def,
                    dependency_structure,
                    handle,
                    parent_step_inputs,
                )

                # If an input with runtime_type "Nothing" doesnt have a value
                # we don't create a StepInput
                if step_input is None:
                    continue

                step_input = create_subplan_for_input(self, solid, step_input, input_def, handle)

                step_inputs.append(step_input)

            ### 2. COMPUTE FUNCTION OR RECURSE
            # Create and add execution plan step for the solid compute function or
            # recurse over the solids in a CompositeSolid
            if isinstance(solid.definition, SolidDefinition):
                solid_transform_step = create_compute_step(
                    self.pipeline_name, self.environment_config, solid, step_inputs, handle
                )
                self.add_step(solid_transform_step)
                terminal_transform_step = solid_transform_step
            elif isinstance(solid.definition, CompositeSolidDefinition):
                terminal_transform_step = self._build_from_sorted_solids(
                    solids_in_topological_order(solid.definition),
                    solid.definition.dependency_structure,
                    parent_handle=handle,
                    parent_step_inputs=step_inputs,
                )
            else:
                check.invariant(
                    False,
                    'Unexpected solid type {type} encountered during execution planning'.format(
                        type=type(solid.definition)
                    ),
                )

            ### 3. OUTPUTS
            # Create and add execution plan steps (and output handles) for solid outputs
            for name, output_def in solid.definition.output_dict.items():
                subplan = create_subplan_for_output(
                    self.pipeline_name,
                    self.environment_config,
                    solid,
                    terminal_transform_step,
                    output_def,
                    handle,
                )
                self.add_steps(subplan.steps)

                output_handle = solid.output_handle(name)
                self.set_output_handle(output_handle, subplan.terminal_step_output_handle)

        return terminal_transform_step


def create_subplan_for_input(plan_builder, solid, step_input, input_def, handle):
    check.inst_param(plan_builder, 'plan_builder', _PlanBuilder)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(step_input, 'step_input', StepInput)
    check.inst_param(input_def, 'input_def', InputDefinition)
    check.inst_param(handle, 'handle', SolidHandle)

    if plan_builder.environment_config.expectations.evaluate and input_def.expectations:
        subplan = create_expectations_subplan(
            plan_builder.pipeline_name,
            solid,
            input_def,
            step_input,
            kind=StepKind.INPUT_EXPECTATION,
            handle=handle,
        )
        plan_builder.add_steps(subplan.steps)
        return StepInput(
            input_def.name, input_def.runtime_type, subplan.terminal_step_output_handle
        )
    else:
        return step_input


def create_subplan_for_output(
    pipeline_name, environment_config, solid, solid_transform_step, output_def, handle
):
    check.str_param(pipeline_name, 'pipeline_name')
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(solid_transform_step, 'solid_transform_step', ExecutionStep)
    check.inst_param(output_def, 'output_def', OutputDefinition)

    return decorate_with_expectations(
        pipeline_name, environment_config, solid, solid_transform_step, output_def, handle
    )


def get_step_input(
    plan_builder, solid, input_name, input_def, dependency_structure, handle, parent_step_inputs
):
    check.inst_param(plan_builder, 'plan_builder', _PlanBuilder)
    check.inst_param(solid, 'solid', Solid)
    check.str_param(input_name, 'input_name')
    check.inst_param(input_def, 'input_def', InputDefinition)
    check.inst_param(dependency_structure, 'dependency_structure', DependencyStructure)
    check.opt_inst_param(handle, 'handle', SolidHandle)
    check.opt_list_param(parent_step_inputs, 'parent_step_inputs', of_type=StepInput)

    solid_config = plan_builder.environment_config.solids.get(str(handle))
    if solid_config and input_name in solid_config.inputs:
        return StepInput(
            input_name, input_def.runtime_type, config_data=solid_config.inputs[input_name]
        )

    input_handle = solid.input_handle(input_name)
    if dependency_structure.has_singular_dep(input_handle):
        solid_output_handle = dependency_structure.get_singular_dep(input_handle)
        return StepInput(
            input_name, input_def.runtime_type, plan_builder.get_output_handle(solid_output_handle)
        )

    if dependency_structure.has_deps(input_handle):
        solid_output_handles = dependency_structure.get_deps(input_handle)
        join_data = create_join_outputs_step(
            plan_builder.pipeline_name,
            input_handle,
            solid_output_handles,
            {
                solid_output_handle: plan_builder.get_output_handle(solid_output_handle)
                for solid_output_handle in solid_output_handles
            },
            handle,
        )
        plan_builder.add_step(join_data.step)
        return StepInput(input_name, input_def.runtime_type, join_data.step_output_handle)

    if solid.container_maps_input(input_name):
        parent_name = solid.container_mapped_input(input_name).definition.name
        parent_inputs = {step_input.name: step_input for step_input in parent_step_inputs}
        if parent_name in parent_inputs:
            parent_input = parent_inputs[parent_name]
            return StepInput(
                input_name,
                input_def.runtime_type,
                parent_input.prev_output_handle,
                parent_input.config_data,
            )

    # At this point we have an input that is not hooked up to
    # the output of another solid or provided via environment config.

    # We will allow this for "Nothing" type inputs and continue.
    if input_def.runtime_type.is_nothing:
        return None

    # Otherwise we throw an error.
    raise DagsterInvariantViolationError(
        (
            'In pipeline {pipeline_name} solid {solid_name}, input {input_name} '
            'must get a value either (a) from a dependency or (b) from the '
            'inputs section of its configuration.'
        ).format(
            pipeline_name=plan_builder.pipeline_name, solid_name=solid.name, input_name=input_name
        )
    )


class ExecutionPlan(
    namedtuple('_ExecutionPlan', 'pipeline_def step_dict deps steps artifacts_persisted')
):
    def __new__(cls, pipeline_def, step_dict, deps, artifacts_persisted):
        return super(ExecutionPlan, cls).__new__(
            cls,
            pipeline_def=check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition),
            step_dict=check.dict_param(
                step_dict, 'step_dict', key_type=str, value_type=ExecutionStep
            ),
            deps=check.dict_param(deps, 'deps', key_type=str, value_type=set),
            steps=list(step_dict.values()),
            artifacts_persisted=check.bool_param(artifacts_persisted, 'artifacts_persisted'),
        )

    def get_step_output(self, step_output_handle):
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        step = self.get_step_by_key(step_output_handle.step_key)
        return step.step_output_named(step_output_handle.output_name)

    def has_step(self, key):
        check.str_param(key, 'key')
        return key in self.step_dict

    def get_step_by_key(self, key):
        check.str_param(key, 'key')
        return self.step_dict[key]

    def topological_steps(self):
        return [step for step_level in self.topological_step_levels() for step in step_level]

    def topological_step_levels(self):
        return [
            [self.step_dict[step_key] for step_key in step_key_level]
            for step_key_level in toposort(self.deps)
        ]

    @staticmethod
    def build(pipeline_def, environment_config, mode_definition):
        '''Here we build a new ExecutionPlan from a pipeline definition and the environment config.

        To do this, we iterate through the pipeline's solids in topological order, and hand off the
        execution steps for each solid to a companion _PlanBuilder object.

        Once we've processed the entire pipeline, we invoke _PlanBuilder.build() to construct the
        ExecutionPlan object.
        '''
        check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
        check.inst_param(environment_config, 'environment_config', EnvironmentConfig)

        plan_builder = _PlanBuilder(pipeline_def, environment_config, mode_definition)

        # Finally, we build and return the execution plan
        return plan_builder.build()



import pytest

from dagster import (
    CompositeSolidDefinition,
    DagsterInvalidDefinitionError,
    DependencyDefinition,
    ExpectationResult,
    Field,
    InputDefinition,
    OutputDefinition,
    PipelineDefinition,
    SolidInvocation,
    String,
    composite_solid,
    dagster_type,
    execute_pipeline,
    lambda_solid,
    pipeline,
    solid,
)
from dagster.core.definitions import IOExpectationDefinition
from dagster.core.utility_solids import (
    create_root_solid,
    create_solid_with_deps,
    define_stub_solid,
    input_set,
)


def test_composite_basic_execution():
    a_source = define_stub_solid('A_source', [input_set('A_input')])
    node_a = create_root_solid('A')
    node_b = create_solid_with_deps('B', node_a)
    node_c = create_solid_with_deps('C', node_a)
    node_d = create_solid_with_deps('D', node_b, node_c)

    diamond_composite = CompositeSolidDefinition(
        name='diamond_composite',
        solid_defs=[a_source, node_a, node_b, node_c, node_d],
        dependencies={
            'A': {'A_input': DependencyDefinition('A_source')},
            'B': {'A': DependencyDefinition('A')},
            'C': {'A': DependencyDefinition('A')},
            'D': {'B': DependencyDefinition('B'), 'C': DependencyDefinition('C')},
        },
    )

    result = execute_pipeline(PipelineDefinition(solid_defs=[diamond_composite]))
    assert result.success

    result = execute_pipeline(
        PipelineDefinition(
            solid_defs=[diamond_composite],
            dependencies={
                SolidInvocation('diamond_composite', alias='D1'): {},
                SolidInvocation('diamond_composite', alias='D2'): {},
            },
        )
    )
    assert result.success

    wrapped_composite = CompositeSolidDefinition(
        name='wrapped_composite', solid_defs=[diamond_composite]
    )
    result = execute_pipeline(PipelineDefinition(solid_defs=[diamond_composite, wrapped_composite]))
    assert result.success

    empty_composite = CompositeSolidDefinition(name='empty', solid_defs=[])
    result = execute_pipeline(PipelineDefinition(solid_defs=[empty_composite]))
    assert result.success


def test_composite_config():
    called = {}

    @solid(config_field=Field(String))
    def configured(context):
        called['configured'] = True
        assert context.solid_config is 'yes'

    inner = CompositeSolidDefinition(name='inner', solid_defs=[configured])
    outer = CompositeSolidDefinition(name='outer', solid_defs=[inner])
    pipe = PipelineDefinition(name='composites_pipeline', solid_defs=[outer])
    result = execute_pipeline(
        pipe,
        {'solids': {'outer': {'solids': {'inner': {'solids': {'configured': {'config': 'yes'}}}}}}},
    )
    assert result.success
    assert called['configured']


def test_composite_config_input():
    called = {}

    @solid(input_defs=[InputDefinition('one')])
    def node_a(_context, one):
        called['node_a'] = True
        assert one is 1

    inner = CompositeSolidDefinition(name='inner', solid_defs=[node_a])
    outer = CompositeSolidDefinition(name='outer', solid_defs=[inner])
    pipe = PipelineDefinition(name='composites_pipeline', solid_defs=[outer])
    result = execute_pipeline(
        pipe,
        {
            'solids': {
                'outer': {
                    'solids': {'inner': {'solids': {'node_a': {'inputs': {'one': {'value': 1}}}}}}
                }
            }
        },
    )
    assert result.success
    assert called['node_a']


def test_mapped_composite_config_input():
    called = {}

    @solid(input_defs=[InputDefinition('one')])
    def node_a(_context, one):
        called['node_a'] = True
        assert one is 1

    inner = CompositeSolidDefinition(
        name='inner',
        solid_defs=[node_a],
        input_mappings=[InputDefinition('inner_one').mapping_to('node_a', 'one')],
    )
    outer = CompositeSolidDefinition(
        name='outer',
        solid_defs=[inner],
        input_mappings=[InputDefinition('outer_one').mapping_to('inner', 'inner_one')],
    )
    pipe = PipelineDefinition(name='composites_pipeline', solid_defs=[outer])

    result = execute_pipeline(pipe, {'solids': {'outer': {'inputs': {'outer_one': {'value': 1}}}}})
    assert result.success
    assert called['node_a']


def test_mapped_composite_input_expectations():
    called = {}

    def exp_a(_c, _v):
        called['exp_a'] = True
        return ExpectationResult(True)

    @solid(
        input_defs=[InputDefinition('one', expectations=[IOExpectationDefinition('exp_a', exp_a)])]
    )
    def node_a(_context, one):
        called['node_a'] = True
        assert one is 1

    def inner_exp(_c, _v):
        called['inner_exp'] = True
        return ExpectationResult(True)

    def outer_exp(_c, _v):
        called['outer_exp'] = True
        return ExpectationResult(True)

    inner = CompositeSolidDefinition(
        name='inner',
        solid_defs=[node_a],
        input_mappings=[
            InputDefinition(
                name='inner_one', expectations=[IOExpectationDefinition('inner_exp', inner_exp)]
            ).mapping_to('node_a', 'one')
        ],
    )
    outer = CompositeSolidDefinition(
        name='outer',
        solid_defs=[inner],
        input_mappings=[
            InputDefinition(
                'outer_one', expectations=[IOExpectationDefinition('outer_exp', outer_exp)]
            ).mapping_to('inner', 'inner_one')
        ],
    )
    pipe = PipelineDefinition(name='composites_pipeline', solid_defs=[outer])

    result = execute_pipeline(pipe, {'solids': {'outer': {'inputs': {'outer_one': {'value': 1}}}}})
    assert result.success
    assert called['node_a']
    assert called['exp_a']
    assert called['inner_exp']
    assert called['outer_exp']


def test_composite_io_mapping():
    a_source = define_stub_solid('A_source', [input_set('A_input')])
    node_a = create_root_solid('A')

    node_b = create_solid_with_deps('B', node_a)
    node_c = create_solid_with_deps('C', node_b)

    comp_a_inner = CompositeSolidDefinition(
        name='comp_a_inner',
        solid_defs=[a_source, node_a],
        dependencies={'A': {'A_input': DependencyDefinition('A_source')}},
        output_mappings=[OutputDefinition().mapping_from('A')],
    )

    comp_a_outer = CompositeSolidDefinition(
        name='comp_a_outer',
        solid_defs=[comp_a_inner],
        output_mappings=[OutputDefinition().mapping_from('comp_a_inner')],
    )

    comp_bc_inner = CompositeSolidDefinition(
        name='comp_bc_inner',
        solid_defs=[node_b, node_c],
        dependencies={'C': {'B': DependencyDefinition('B')}},
        input_mappings=[
            InputDefinition(name='inner_B_in').mapping_to(solid_name='B', input_name='A')
        ],
    )

    comp_bc_outer = CompositeSolidDefinition(
        name='comp_bc_outer',
        solid_defs=[comp_bc_inner],
        dependencies={},
        input_mappings=[
            InputDefinition(name='outer_B_in').mapping_to(
                solid_name='comp_bc_inner', input_name='inner_B_in'
            )
        ],
    )
    result = execute_pipeline(
        PipelineDefinition(
            name='wrapped_io',
            solid_defs=[comp_a_outer, comp_bc_outer],
            dependencies={'comp_bc_outer': {'outer_B_in': DependencyDefinition('comp_a_outer')}},
        )
    )
    assert result.success


def test_io_error_is_decent():
    with pytest.raises(DagsterInvalidDefinitionError, match='mapping_to'):
        CompositeSolidDefinition(
            name='comp_a_outer', solid_defs=[], input_mappings=[InputDefinition('should_be_mapped')]
        )

    with pytest.raises(DagsterInvalidDefinitionError, match='mapping_from'):
        CompositeSolidDefinition(
            name='comp_a_outer', solid_defs=[], output_mappings=[OutputDefinition()]
        )


def test_types_descent():
    @dagster_type
    class Foo(object):
        pass

    @solid(output_defs=[OutputDefinition(Foo)])
    def inner_solid(_context):
        return Foo()

    middle_solid = CompositeSolidDefinition(name='middle_solid', solid_defs=[inner_solid])

    outer_solid = CompositeSolidDefinition(name='outer_solid', solid_defs=[middle_solid])

    pipe = PipelineDefinition(name='layered_types', solid_defs=[outer_solid])

    assert pipe.has_runtime_type('Foo')


def test_deep_mapping():
    @lambda_solid
    def echo(blah):
        return blah

    @lambda_solid
    def emit_foo():
        return 'foo'

    @composite_solid(output_defs=[OutputDefinition(String, 'z')])
    def az(a):
        return echo(a)

    @composite_solid(output_defs=[OutputDefinition(String, 'y')])
    def by(b):
        return az(b)

    @composite_solid(output_defs=[OutputDefinition(String, 'x')])
    def cx(c):
        return by(c)

    @pipeline
    def nested():
        echo(cx(emit_foo()))

    result = execute_pipeline(nested)
    assert result.result_for_solid('echo').result_value() == 'foo'



from collections import defaultdict, OrderedDict
import itertools

from dagster import check

from dagster.core.definitions import PipelineDefinition, Solid
from dagster.core.definitions.utils import DEFAULT_OUTPUT
from dagster.core.errors import DagsterInvariantViolationError
from dagster.core.events import DagsterEvent, DagsterEventType
from dagster.core.execution.plan.objects import StepKind


class PipelineExecutionResult(object):
    '''Output of execution of the whole pipeline. Returned eg by :py:func:`execute_pipeline`.
    '''

    def __init__(self, pipeline, run_id, event_list, reconstruct_context):
        self.pipeline = check.inst_param(pipeline, 'pipeline', PipelineDefinition)
        self.run_id = check.str_param(run_id, 'run_id')
        self.event_list = check.list_param(event_list, 'step_event_list', of_type=DagsterEvent)
        self.reconstruct_context = check.callable_param(reconstruct_context, 'reconstruct_context')

        # Dict[string: solid_name, SolidExecutionResult]
        solid_result_dict = self._context_solid_result_dict(event_list)

        self.solid_result_dict = solid_result_dict
        self.solid_result_list = list(self.solid_result_dict.values())

    def _context_solid_result_dict(self, event_list):
        solid_set = set()
        solid_order = []
        step_events_by_solid_by_kind = defaultdict(lambda: defaultdict(list))
        step_event_list = [event for event in event_list if event.is_step_event]
        for step_event in step_event_list:
            solid_handle = step_event.solid_handle
            if solid_handle not in solid_set:
                solid_order.append(solid_handle)
                solid_set.add(solid_handle)

            step_events_by_solid_by_kind[solid_handle][step_event.step_kind].append(step_event)

        solid_result_dict = OrderedDict()

        for solid_handle in solid_order:
            solid_result_dict[solid_handle] = SolidExecutionResult(
                self.pipeline.get_solid(solid_handle),
                dict(step_events_by_solid_by_kind[solid_handle]),
                self.reconstruct_context,
            )
        return solid_result_dict

    @property
    def success(self):
        '''Whether the pipeline execution was successful at all steps'''
        return all([not event.is_failure for event in self.event_list])

    @property
    def step_event_list(self):
        '''The full list of step events'''
        return [event for event in self.event_list if event.is_step_event]

    def result_for_solid(self, name):
        '''Get a :py:class:`SolidExecutionResult` for a given top level solid name.
        '''
        check.str_param(name, 'name')

        if not self.pipeline.has_solid_named(name):
            raise DagsterInvariantViolationError(
                'Try to get result for solid {name} in {pipeline}. No such solid.'.format(
                    name=name, pipeline=self.pipeline.display_name
                )
            )
        top_level_results = {
            solid_id.name: result
            for solid_id, result in self.solid_result_dict.items()
            if solid_id.parent is None
        }
        if name not in top_level_results:
            raise DagsterInvariantViolationError(
                'Did not find result for solid {name} in pipeline execution result'.format(
                    name=name
                )
            )

        return top_level_results[name]

    def result_for_handle(self, handle):
        '''Get a :py:class:`SolidExecutionResult` for a given solid handle string.
        '''
        check.str_param(handle, 'handle')
        results = {
            solid_id.to_string(): result for solid_id, result in self.solid_result_dict.items()
        }

        if handle not in results:
            raise DagsterInvariantViolationError(
                'Did not find result for solid handle {handle} in pipeline execution result'.format(
                    handle=handle
                )
            )

        return results[handle]


class SolidExecutionResult(object):
    '''Execution result for one solid of the pipeline.
    '''

    def __init__(self, solid, step_events_by_kind, reconstruct_context):
        self.solid = check.inst_param(solid, 'solid', Solid)
        self.step_events_by_kind = check.dict_param(
            step_events_by_kind, 'step_events_by_kind', key_type=StepKind, value_type=list
        )
        self.reconstruct_context = check.callable_param(reconstruct_context, 'reconstruct_context')

    @property
    def compute_input_event_dict(self):
        '''input_events_during_compute keyed by input name'''
        return {se.event_specific_data.input_name: se for se in self.input_events_during_compute}

    @property
    def input_events_during_compute(self):
        '''All events of type STEP_INPUT'''
        return self._compute_steps_of_type(DagsterEventType.STEP_INPUT)

    @property
    def compute_output_event_dict(self):
        '''output_events_during_compute keyed by output name'''
        return {se.event_specific_data.output_name: se for se in self.output_events_during_compute}

    def get_output_event_for_compute(self, output_name='result'):
        '''The STEP_OUTPUT events for the given output name'''
        return self.compute_output_event_dict[output_name]

    @property
    def output_events_during_compute(self):
        '''All events of type STEP_OUTPUT'''
        return self._compute_steps_of_type(DagsterEventType.STEP_OUTPUT)

    @property
    def compute_step_events(self):
        '''All events that happen during the solid compute function'''
        return self.step_events_by_kind.get(StepKind.COMPUTE, [])

    @property
    def materializations_during_compute(self):
        '''The Materializations objects yielded by the solid'''
        return [
            mat_event.event_specific_data.materialization
            for mat_event in self.materialization_events_during_compute
        ]

    @property
    def materialization_events_during_compute(self):
        '''All events of type STEP_MATERIALIZATION'''
        return self._compute_steps_of_type(DagsterEventType.STEP_MATERIALIZATION)

    @property
    def expectation_events_during_compute(self):
        '''All events of type STEP_EXPECTATION_RESULT'''
        return self._compute_steps_of_type(DagsterEventType.STEP_EXPECTATION_RESULT)

    def _compute_steps_of_type(self, dagster_event_type):
        return list(
            filter(lambda se: se.event_type == dagster_event_type, self.compute_step_events)
        )

    @property
    def expectation_results_during_compute(self):
        '''The ExpectationResult objects yielded by the solid'''
        return [
            expt_event.event_specific_data.expectation_result
            for expt_event in self.expectation_events_during_compute
        ]

    def get_step_success_event(self):
        '''The STEP_SUCCESS event, throws if not present'''
        for step_event in self.compute_step_events:
            if step_event.event_type == DagsterEventType.STEP_SUCCESS:
                return step_event

        check.failed('Step success not found for solid {}'.format(self.solid.name))

    @property
    def input_expectation_step_events(self):
        '''All events of type INPUT_EXPECTATION'''
        return self.step_events_by_kind.get(StepKind.INPUT_EXPECTATION, [])

    @property
    def output_expectation_step_events(self):
        '''All events of type OUTPUT_EXPECTATION'''
        return self.step_events_by_kind.get(StepKind.OUTPUT_EXPECTATION, [])

    @property
    def compute_step_failure_event(self):
        '''The STEP_FAILURE event, throws if it did not fail'''
        if self.success:
            raise DagsterInvariantViolationError(
                'Cannot call compute_step_failure_event if successful'
            )

        step_failure_events = self._compute_steps_of_type(DagsterEventType.STEP_FAILURE)
        check.invariant(len(step_failure_events) == 1)
        return step_failure_events[0]

    @property
    def success(self):
        '''Whether the solid execution was successful'''
        any_success = False
        for step_event in itertools.chain(
            self.input_expectation_step_events,
            self.output_expectation_step_events,
            self.compute_step_events,
        ):
            if step_event.event_type == DagsterEventType.STEP_FAILURE:
                return False
            if step_event.event_type == DagsterEventType.STEP_SUCCESS:
                any_success = True

        return any_success

    @property
    def skipped(self):
        '''Whether the solid execution was skipped'''
        return all(
            [
                step_event.event_type == DagsterEventType.STEP_SKIPPED
                for step_event in itertools.chain(
                    self.input_expectation_step_events,
                    self.output_expectation_step_events,
                    self.compute_step_events,
                )
            ]
        )

    @property
    def result_values(self):
        '''Return dictionary of transformed results, with keys being output names.
        Returns None if execution result isn't a success.

        Reconstructs the pipeline context to materialize values.
        '''
        if self.success and self.compute_step_events:
            with self.reconstruct_context() as context:
                values = {
                    compute_step_event.step_output_data.output_name: self._get_value(
                        context, compute_step_event.step_output_data
                    )
                    for compute_step_event in self.compute_step_events
                    if compute_step_event.is_successful_output
                }
            return values
        else:
            return None

    def result_value(self, output_name=DEFAULT_OUTPUT):
        '''Returns transformed value either for DEFAULT_OUTPUT or for the output
        given as output_name. Returns None if execution result isn't a success.

        Reconstructs the pipeline context to materialize value.
        '''
        check.str_param(output_name, 'output_name')

        if not self.solid.definition.has_output(output_name):
            raise DagsterInvariantViolationError(
                '{output_name} not defined in solid {solid}'.format(
                    output_name=output_name, solid=self.solid.name
                )
            )

        if self.success:
            for compute_step_event in self.compute_step_events:
                if (
                    compute_step_event.is_successful_output
                    and compute_step_event.step_output_data.output_name == output_name
                ):
                    with self.reconstruct_context() as context:
                        value = self._get_value(context, compute_step_event.step_output_data)
                    return value

            raise DagsterInvariantViolationError(
                (
                    'Did not find result {output_name} in solid {self.solid.name} '
                    'execution result'
                ).format(output_name=output_name, self=self)
            )
        else:
            return None

    def _get_value(self, context, step_output_data):
        return context.intermediates_manager.get_intermediate(
            context=context,
            runtime_type=self.solid.output_def_named(step_output_data.output_name).runtime_type,
            step_output_handle=step_output_data.step_output_handle,
        )

    @property
    def failure_data(self):
        '''Returns the failing step's data that happened during this solid's execution, if any'''
        for step_event in itertools.chain(
            self.input_expectation_step_events,
            self.output_expectation_step_events,
            self.compute_step_events,
        ):
            if step_event.event_type == DagsterEventType.STEP_FAILURE:
                return step_event.step_failure_data



import pytest

from dagster import (
    DagsterInvariantViolationError,
    DependencyDefinition,
    ExecutionTargetHandle,
    InProcessExecutorConfig,
    InputDefinition,
    MultiprocessExecutorConfig,
    PipelineDefinition,
    RunConfig,
    execute_pipeline,
    lambda_solid,
)


def test_diamond_simple_execution():
    result = execute_pipeline(
        define_diamond_pipeline(), run_config=RunConfig(executor_config=InProcessExecutorConfig())
    )
    assert result.success
    assert result.result_for_solid('adder').result_value() == 11


def transform_event(result, solid_name):
    return result.result_for_solid(solid_name).compute_step_events[0]


def test_diamond_multi_execution():
    pipeline = define_diamond_pipeline()
    result = execute_pipeline(
        pipeline,
        environment_dict={'storage': {'filesystem': {}}},
        run_config=RunConfig(
            executor_config=MultiprocessExecutorConfig(
                ExecutionTargetHandle.for_pipeline_fn(define_diamond_pipeline)
            )
        ),
    )
    assert result.success

    assert result.result_for_solid('adder').result_value() == 11

    pids_by_solid = {}
    for solid in pipeline.solids:
        pids_by_solid[solid.name] = transform_event(result, solid.name).logging_tags['pid']

    # guarantee that all solids ran in their own process
    assert len(set(pids_by_solid.values())) == len(pipeline.solids)


def define_diamond_pipeline():
    @lambda_solid
    def return_two():
        return 2

    @lambda_solid(input_defs=[InputDefinition('num')])
    def add_three(num):
        return num + 3

    @lambda_solid(input_defs=[InputDefinition('num')])
    def mult_three(num):
        return num * 3

    @lambda_solid(input_defs=[InputDefinition('left'), InputDefinition('right')])
    def adder(left, right):
        return left + right

    return PipelineDefinition(
        name='diamond_execution',
        solid_defs=[return_two, add_three, mult_three, adder],
        dependencies={
            'add_three': {'num': DependencyDefinition('return_two')},
            'mult_three': {'num': DependencyDefinition('return_two')},
            'adder': {
                'left': DependencyDefinition('add_three'),
                'right': DependencyDefinition('mult_three'),
            },
        },
    )


def define_error_pipeline():
    @lambda_solid
    def throw_error():
        raise Exception('bad programmer')

    return PipelineDefinition(name='error_pipeline', solid_defs=[throw_error])


def test_error_pipeline():
    pipeline = define_error_pipeline()
    result = execute_pipeline(
        pipeline,
        run_config=RunConfig(executor_config=InProcessExecutorConfig(raise_on_error=False)),
    )
    assert not result.success


def test_error_pipeline_multiprocess():
    pipeline = define_error_pipeline()
    result = execute_pipeline(
        pipeline,
        run_config=RunConfig(
            executor_config=MultiprocessExecutorConfig(
                ExecutionTargetHandle.for_pipeline_fn(define_error_pipeline)
            )
        ),
        environment_dict={'storage': {'filesystem': {}}},
    )
    assert not result.success


def test_mem_storage_error_pipeline_multiprocess():
    with pytest.raises(DagsterInvariantViolationError) as exc_info:
        execute_pipeline(
            define_diamond_pipeline(),
            run_config=RunConfig(
                executor_config=MultiprocessExecutorConfig(
                    ExecutionTargetHandle.for_pipeline_fn(define_error_pipeline)
                )
            ),
        )

    assert (
        'While invoking '
        'pipeline diamond_execution. You have attempted to use the '
        'multiprocessing executor while using system storage in_memory '
        'which does not persist intermediates. This means there would '
        'be no way to move data between different processes. Please '
        'configure your pipeline in the storage config section to use '
        'persistent system storage such as the filesystem.'
    ) in str(exc_info.value)



from dagster import check
from dagster.core.definitions import (
    DagsterIOExpectationFailedError,
    ExpectationResult,
    InputDefinition,
    IOExpectationDefinition,
    Output,
    OutputDefinition,
    Solid,
    SolidHandle,
)
from dagster.core.errors import DagsterInvariantViolationError
from dagster.core.execution.context.expectation import ExpectationExecutionContext
from dagster.core.execution.context.system import SystemStepExecutionContext
from dagster.core.system_config.objects import EnvironmentConfig

from .objects import (
    ExecutionStep,
    ExecutionValueSubplan,
    StepInput,
    StepKind,
    StepOutput,
    StepOutputHandle,
)
from .utility import create_joining_subplan

EXPECTATION_INPUT = 'expectation_input'
EXPECTATION_VALUE_OUTPUT = 'expectation_value'


def _create_expectation_lambda(solid, inout_def, expectation_def, internal_output_name):
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(inout_def, 'inout_def', (InputDefinition, OutputDefinition))
    check.inst_param(expectation_def, 'expectations_def', IOExpectationDefinition)
    check.str_param(internal_output_name, 'internal_output_name')

    def _do_expectation(expectation_context, inputs):
        check.inst_param(expectation_context, 'step_context', SystemStepExecutionContext)
        value = inputs[EXPECTATION_INPUT]
        expectation_context = expectation_context.for_expectation(inout_def, expectation_def)
        expt_result = expectation_def.expectation_fn(
            ExpectationExecutionContext(expectation_context), value
        )

        if not isinstance(expt_result, ExpectationResult):
            raise DagsterInvariantViolationError(
                (
                    'Expectation for solid {solid_name} on {desc_key} {inout_name} '
                    'did not return an ExpectationResult'.format(
                        solid_name=solid.name,
                        desc_key=inout_def.descriptive_key,
                        inout_name=inout_def.name,
                    )
                )
            )

        if expt_result.success:
            yield expt_result
            yield Output(output_name=internal_output_name, value=inputs[EXPECTATION_INPUT])
        else:
            expectation_context.log.debug(
                'Expectation {key} failed on {value}.'.format(
                    key=expectation_context.step.key, value=value
                )
            )
            raise DagsterIOExpectationFailedError(expectation_context, value)

    return _do_expectation


def create_expectations_subplan(pipeline_name, solid, inout_def, step_input, kind, handle):
    check.str_param(pipeline_name, 'pipeline_name')
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(inout_def, 'inout_def', (InputDefinition, OutputDefinition))
    check.inst_param(step_input, 'step_input', StepInput)
    check.inst_param(kind, 'kind', StepKind)

    input_expect_steps = []
    for expectation_def in inout_def.expectations:
        expect_step = create_expectation_step(
            pipeline_name=pipeline_name,
            solid=solid,
            expectation_def=expectation_def,
            key_suffix='{desc_key}.{inout_name}.expectation.{expectation_name}'.format(
                desc_key=inout_def.descriptive_key,
                inout_name=inout_def.name,
                expectation_name=expectation_def.name,
            ),
            kind=kind,
            step_input=step_input,
            inout_def=inout_def,
            handle=handle,
        )
        input_expect_steps.append(expect_step)

    return create_joining_subplan(
        pipeline_name,
        solid,
        '{desc_key}.{inout_name}.expectations.join'.format(
            desc_key=inout_def.descriptive_key, inout_name=inout_def.name
        ),
        input_expect_steps,
        EXPECTATION_VALUE_OUTPUT,
        handle,
    )


def create_expectation_step(
    pipeline_name, solid, expectation_def, key_suffix, kind, step_input, inout_def, handle
):
    check.str_param(pipeline_name, 'pipeline_name')
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(expectation_def, 'expectation_def', IOExpectationDefinition)
    check.str_param(key_suffix, 'key_suffix')
    check.inst_param(kind, 'kind', StepKind)
    check.inst_param(step_input, 'step_input', StepInput)
    check.inst_param(inout_def, 'inout_def', (InputDefinition, OutputDefinition))
    check.opt_inst_param(handle, 'handle', SolidHandle)

    value_type = inout_def.runtime_type

    return ExecutionStep(
        pipeline_name=pipeline_name,
        key_suffix=key_suffix,
        step_inputs=[
            StepInput(
                name=EXPECTATION_INPUT,
                runtime_type=value_type,
                prev_output_handle=step_input.prev_output_handle,
                config_data=step_input.config_data,
            )
        ],
        step_outputs=[
            # Expectation value output is optional since we omit if the expectation fails
            StepOutput(name=EXPECTATION_VALUE_OUTPUT, runtime_type=value_type, optional=True)
        ],
        compute_fn=_create_expectation_lambda(
            solid, inout_def, expectation_def, EXPECTATION_VALUE_OUTPUT
        ),
        kind=kind,
        solid_handle=handle,
        logging_tags={
            'expectation': expectation_def.name,
            inout_def.descriptive_key: inout_def.name,
        },
    )


def decorate_with_expectations(
    pipeline_name, environment_config, solid, transform_step, output_def, handle
):
    check.str_param(pipeline_name, 'pipeline_name')
    check.inst_param(environment_config, 'environment_config', EnvironmentConfig)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(transform_step, 'transform_step', ExecutionStep)
    check.inst_param(output_def, 'output_def', OutputDefinition)
    check.opt_inst_param(handle, 'handle', SolidHandle)

    # resolve to the output def from the transform_step's solid
    output_def = solid.definition.resolve_output_to_origin(output_def.name)
    terminal_step_output_handle = StepOutputHandle.from_step(transform_step, output_def.name)

    if environment_config.expectations.evaluate and output_def.expectations:
        return create_expectations_subplan(
            pipeline_name,
            solid,
            output_def,
            StepInput(output_def.name, output_def.runtime_type, terminal_step_output_handle),
            kind=StepKind.OUTPUT_EXPECTATION,
            handle=handle,
        )
    else:
        return ExecutionValueSubplan.empty(terminal_step_output_handle)



from dagster import check
from dagster.core.errors import DagsterInvalidDefinitionError, DagsterInvariantViolationError
from dagster.core.execution.config import RunConfig
from dagster.core.types.runtime import construct_runtime_type_dictionary

from .container import IContainSolids, create_execution_structure, validate_dependency_dict
from .dependency import (
    DependencyDefinition,
    MultiDependencyDefinition,
    SolidHandle,
    SolidInvocation,
)
from .mode import ModeDefinition
from .preset import PresetDefinition
from .solid import ISolidDefinition


def _check_solids_arg(pipeline_name, solid_defs):
    if not isinstance(solid_defs, list):
        raise DagsterInvalidDefinitionError(
            '"solids" arg to pipeline "{name}" is not a list. Got {val}.'.format(
                name=pipeline_name, val=repr(solid_defs)
            )
        )
    for solid_def in solid_defs:
        if isinstance(solid_def, ISolidDefinition):
            continue
        elif callable(solid_def):
            raise DagsterInvalidDefinitionError(
                '''You have passed a lambda or function {func} into pipeline {name} that is
                not a solid. You have likely forgetten to annotate this function with
                an @solid or @lambda_solid decorator.'
                '''.format(
                    name=pipeline_name, func=solid_def.__name__
                )
            )
        else:
            raise DagsterInvalidDefinitionError(
                'Invalid item in solid list: {item}'.format(item=repr(solid_def))
            )

    return solid_defs


class PipelineDefinition(IContainSolids, object):
    '''A instance of a PipelineDefinition represents a pipeline in dagster.

    A pipeline is comprised of:

    - Solids:
        Each solid represents a functional unit of data computation.

    - Dependencies:
        Solids within a pipeline are arranged as a DAG (directed, acyclic graph). Dependencies
        determine how the values produced by solids flow through the DAG.

    Args:
        solid_defs (List[SolidDefinition]):
            The set of solid definitions used in this pipeline.
        name (Optional[str])
        description (Optional[str])
        dependencies (Optional[Dict[Union[str, SolidInvocation], Dict[str, DependencyDefinition]]]):
            A structure that declares where each solid gets its inputs. The keys at the top
            level dict are either string names of solids or SolidInvocations. The values
            are dicts that map input names to DependencyDefinitions.
        mode_defs (Optional[List[ModeDefinition]]):
            The set of modes this pipeline can operate in. Modes can be used for example to vary
            resources and logging implementations for local testing and running in production.
        preset_defs (Optional[List[PresetDefinition]]):
            Given the different ways a pipeline may execute, presets give you a way to provide
            specific valid collections of configuration.
    '''

    def __init__(
        self,
        solid_defs,
        name=None,
        description=None,
        dependencies=None,
        mode_defs=None,
        preset_defs=None,
    ):
        self.name = check.opt_str_param(name, 'name', '<<unnamed>>')
        self.description = check.opt_str_param(description, 'description')

        mode_definitions = check.opt_list_param(mode_defs, 'mode_defs', of_type=ModeDefinition)

        if not mode_definitions:
            mode_definitions = [ModeDefinition()]

        self.mode_definitions = mode_definitions

        current_level_solid_defs = check.list_param(
            _check_solids_arg(self.name, solid_defs), 'solid_defs', of_type=ISolidDefinition
        )

        seen_modes = set()
        for mode_def in mode_definitions:
            if mode_def.name in seen_modes:
                raise DagsterInvalidDefinitionError(
                    (
                        'Two modes seen with the name "{mode_name}" in "{pipeline_name}". '
                        'Modes must have unique names.'
                    ).format(mode_name=mode_def.name, pipeline_name=self.name)
                )
            seen_modes.add(mode_def.name)

        self.dependencies = validate_dependency_dict(dependencies)

        dependency_structure, pipeline_solid_dict = create_execution_structure(
            current_level_solid_defs, self.dependencies, container_definition=None
        )

        self._solid_dict = pipeline_solid_dict
        self._dependency_structure = dependency_structure

        self._runtime_type_dict = construct_runtime_type_dictionary(current_level_solid_defs)

        self._preset_dict = {}
        for preset in check.opt_list_param(preset_defs, 'preset_defs', PresetDefinition):
            if preset.name in self._preset_dict:
                raise DagsterInvalidDefinitionError(
                    (
                        'Two PresetDefinitions seen with the name "{name}" in "{pipeline_name}". '
                        'PresetDefinitions must have unique names.'
                    ).format(name=preset.name, pipeline_name=self.name)
                )
            if preset.mode not in seen_modes:
                raise DagsterInvalidDefinitionError(
                    (
                        'PresetDefinition "{name}" in "{pipeline_name}" '
                        'references mode "{mode}" which is not defined.'
                    ).format(name=preset.name, pipeline_name=self.name, mode=preset.mode)
                )
            self._preset_dict[preset.name] = preset

        # Validate solid resource dependencies
        _validate_resource_dependencies(self.mode_definitions, current_level_solid_defs)

        self._all_solid_defs = {}
        for current_level_solid_def in current_level_solid_defs:
            for solid_def in current_level_solid_def.iterate_solid_defs():
                self._all_solid_defs[solid_def.name] = solid_def

    def _get_mode_definition(self, mode):
        check.str_param(mode, 'mode')
        for mode_definition in self.mode_definitions:
            if mode_definition.name == mode:
                return mode_definition

        return None

    def get_default_mode(self):
        return self.mode_definitions[0]

    @property
    def is_single_mode(self):
        return len(self.mode_definitions) == 1

    @property
    def is_multi_mode(self):
        return len(self.mode_definitions) > 1

    def has_mode_definition(self, mode):
        check.str_param(mode, 'mode')
        return bool(self._get_mode_definition(mode))

    def get_default_mode_name(self):
        return self.mode_definitions[0].name

    def get_mode_definition(self, mode=None):
        check.opt_str_param(mode, 'mode')
        if mode is None:
            check.invariant(self.is_single_mode)
            return self.get_default_mode()

        mode_def = self._get_mode_definition(mode)

        if mode_def is None:
            check.failed(
                'Could not find mode {mode} in pipeline {name}'.format(mode=mode, name=self.name)
            )

        return mode_def

    @property
    def available_modes(self):
        return [mode_def.name for mode_def in self.mode_definitions]

    @property
    def display_name(self):
        '''Name suitable for exception messages, logging etc. If pipeline
        is unnamed the method with return "<<unnamed>>".

        Returns:
            str: Display name of pipeline
        '''
        return self.name if self.name else '<<unnamed>>'

    @property
    def solids(self):
        '''Return the top level solids in the pipeline.

        Returns:
            List[Solid]: List of solids.
        '''
        return list(set(self._solid_dict.values()))

    def has_solid_named(self, name):
        '''Return whether or not there is a top level solid with this name in the piepline

        Args:
            name (str): Name of solid

        Returns:
            bool: True if the solid is in the pipeline
        '''
        check.str_param(name, 'name')
        return name in self._solid_dict

    def solid_named(self, name):
        '''Return the top level solid named "name". Throws if it does not exist.

        Args:
            name (str): Name of solid

        Returns:
            Solid:
        '''
        check.str_param(name, 'name')
        if name not in self._solid_dict:
            raise DagsterInvariantViolationError(
                'Pipeline {pipeline_name} has no solid named {name}.'.format(
                    pipeline_name=self.name, name=name
                )
            )
        return self._solid_dict[name]

    def get_solid(self, handle):
        '''Return the solid contained anywhere within the pipeline via its handle.

        Args:
            handle (SolidHandle):

        Returns:
            Solid:

        '''
        check.inst_param(handle, 'handle', SolidHandle)
        current = handle
        lineage = []
        while current:
            lineage.append(current.name)
            current = current.parent

        name = lineage.pop()
        solid = self.solid_named(name)
        while lineage:
            name = lineage.pop()
            solid = solid.definition.solid_named(name)

        return solid

    @property
    def dependency_structure(self):
        return self._dependency_structure

    def has_runtime_type(self, name):
        check.str_param(name, 'name')
        return name in self._runtime_type_dict

    def runtime_type_named(self, name):
        check.str_param(name, 'name')
        return self._runtime_type_dict[name]

    def all_runtime_types(self):
        return self._runtime_type_dict.values()

    @property
    def solid_defs(self):
        return list(self._all_solid_defs.values())

    def solid_def_named(self, name):
        check.str_param(name, 'name')

        check.invariant(name in self._all_solid_defs, '{} not found'.format(name))
        return self._all_solid_defs[name]

    def has_solid_def(self, name):
        check.str_param(name, 'name')
        return name in self._all_solid_defs

    def build_sub_pipeline(self, solid_subset):
        return self if solid_subset is None else _build_sub_pipeline(self, solid_subset)

    def get_presets(self):
        return list(self._preset_dict.values())

    def get_preset(self, name):
        check.str_param(name, 'name')
        if name not in self._preset_dict:
            raise DagsterInvariantViolationError(
                (
                    'Could not find preset for "{name}". Available presets '
                    'for pipeline "{pipeline_name}" are {preset_names}.'
                ).format(
                    name=name, preset_names=list(self._preset_dict.keys()), pipeline_name=self.name
                )
            )

        preset = self._preset_dict[name]

        pipeline = self
        if preset.solid_subset is not None:
            pipeline = pipeline.build_sub_pipeline(preset.solid_subset)

        return {
            'pipeline': pipeline,
            'environment_dict': preset.get_environment_dict(self.name),
            'run_config': RunConfig(mode=preset.mode),
        }


def _dep_key_of(solid):
    return SolidInvocation(solid.definition.name, solid.name)


def _build_sub_pipeline(pipeline_def, solid_names):
    '''
    Build a pipeline which is a subset of another pipeline.
    Only includes the solids which are in solid_names.
    '''

    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.list_param(solid_names, 'solid_names', of_type=str)

    solid_name_set = set(solid_names)
    solids = list(map(pipeline_def.solid_named, solid_names))
    deps = {_dep_key_of(solid): {} for solid in solids}

    def _out_handle_of_inp(input_handle):
        if pipeline_def.dependency_structure.has_deps(input_handle):
            output_handles = pipeline_def.dependency_structure.get_deps(input_handle)
            return [
                output_handle
                for output_handle in output_handles
                if output_handle.solid.name in solid_name_set
            ]
        return []

    for solid in solids:
        for input_handle in solid.input_handles():
            output_handles = _out_handle_of_inp(input_handle)
            if output_handles:
                inner_dep = (
                    DependencyDefinition(
                        solid=output_handles[0].solid.name, output=output_handles[0].output_def.name
                    )
                    if len(output_handles) == 1
                    else MultiDependencyDefinition(
                        [
                            DependencyDefinition(
                                solid=output_handle.solid.name, output=output_handle.output_def.name
                            )
                            for output_handle in output_handles
                        ]
                    )
                )
                deps[_dep_key_of(solid)][input_handle.input_def.name] = inner_dep

    return PipelineDefinition(
        name=pipeline_def.name,
        solid_defs=list({solid.definition for solid in solids}),
        mode_defs=pipeline_def.mode_definitions,
        dependencies=deps,
    )


def _validate_resource_dependencies(mode_definitions, solid_defs):
    '''This validation ensures that each pipeline context provides the resources that are required
    by each solid.
    '''
    check.list_param(mode_definitions, 'mode_definintions', of_type=ModeDefinition)
    check.list_param(solid_defs, 'solid_defs', of_type=ISolidDefinition)

    for mode_def in mode_definitions:
        mode_resources = set(mode_def.resource_defs.keys())
        for solid in solid_defs:
            for required_resource in solid.required_resource_keys:
                if required_resource not in mode_resources:
                    raise DagsterInvalidDefinitionError(
                        (
                            'Resource "{resource}" is required by solid {solid_name}, but is not '
                            'provided by mode "{mode_name}".'
                        ).format(
                            resource=required_resource,
                            solid_name=solid.name,
                            mode_name=mode_def.name,
                        )
                    )
        for system_storage_def in mode_def.system_storage_defs:
            for required_resource in system_storage_def.required_resource_keys:
                if required_resource not in mode_resources:
                    raise DagsterInvalidDefinitionError(
                        (
                            'Resource "{resource}" is required by system storage'
                            '{storage_name}, but is not provided by mode "{mode_name}".'
                        ).format(
                            resource=required_resource,
                            storage_name=system_storage_def.name,
                            mode_name=mode_def.name,
                        )
                    )



'''
This module contains the execution context objects that are internal to the system.
Not every property on these should be exposed to random Jane or Joe dagster user
so we have a different layer of objects that encode the explicit public API
in the user_context module
'''
from collections import namedtuple

from dagster import check
from dagster.utils import merge_dicts

from dagster.core.definitions.expectation import IOExpectationDefinition
from dagster.core.definitions.handle import ExecutionTargetHandle
from dagster.core.definitions.input import InputDefinition
from dagster.core.definitions.mode import ModeDefinition
from dagster.core.definitions.output import OutputDefinition
from dagster.core.definitions.resource import ScopedResourcesBuilder
from dagster.core.log_manager import DagsterLogManager

from dagster.core.storage.file_manager import FileManager
from dagster.core.storage.runs import RunStorage
from dagster.core.system_config.objects import EnvironmentConfig

from ..config import RunConfig


class SystemPipelineExecutionContextData(
    namedtuple(
        '_SystemPipelineExecutionContextData',
        (
            'run_config scoped_resources_builder environment_config pipeline_def '
            'mode_def system_storage_def run_storage intermediates_manager file_manager '
            'execution_target_handle'
        ),
    )
):
    '''
    SystemPipelineExecutionContextData is the data that remains constant throughout the entire
    execution of a pipeline.
    '''

    def __new__(
        cls,
        run_config,
        scoped_resources_builder,
        environment_config,
        pipeline_def,
        mode_def,
        system_storage_def,
        run_storage,
        intermediates_manager,
        file_manager,
        execution_target_handle,
    ):
        from dagster.core.definitions import PipelineDefinition
        from dagster.core.definitions.system_storage import SystemStorageDefinition
        from dagster.core.storage.intermediates_manager import IntermediatesManager

        return super(SystemPipelineExecutionContextData, cls).__new__(
            cls,
            run_config=check.inst_param(run_config, 'run_config', RunConfig),
            scoped_resources_builder=check.inst_param(
                scoped_resources_builder, 'scoped_resources_builder', ScopedResourcesBuilder
            ),
            environment_config=check.inst_param(
                environment_config, 'environment_config', EnvironmentConfig
            ),
            pipeline_def=check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition),
            mode_def=check.inst_param(mode_def, 'mode_def', ModeDefinition),
            system_storage_def=check.inst_param(
                system_storage_def, 'system_storage_def', SystemStorageDefinition
            ),
            run_storage=check.inst_param(run_storage, 'run_storage', RunStorage),
            intermediates_manager=check.inst_param(
                intermediates_manager, 'intermediates_manager', IntermediatesManager
            ),
            # TODO: Make required when https://github.com/dagster-io/dagster/issues/1456
            # is complete
            file_manager=check.opt_inst_param(file_manager, 'file_manager', FileManager),
            execution_target_handle=check.opt_inst_param(
                execution_target_handle, 'execution_target_handle', ExecutionTargetHandle
            ),
        )

    @property
    def run_id(self):
        return self.run_config.run_id

    @property
    def event_callback(self):
        return self.run_config.event_callback

    @property
    def environment_dict(self):
        return self.environment_config.original_config_dict


class SystemPipelineExecutionContext(object):
    __slots__ = ['_pipeline_context_data', '_log_manager', '_legacy_context', '_events']

    def __init__(self, pipeline_context_data, log_manager):
        self._pipeline_context_data = check.inst_param(
            pipeline_context_data, 'pipeline_context_data', SystemPipelineExecutionContextData
        )
        self._log_manager = check.inst_param(log_manager, 'log_manager', DagsterLogManager)

    def for_step(self, step):
        from dagster.core.execution.plan.objects import ExecutionStep

        check.inst_param(step, 'step', ExecutionStep)

        return SystemStepExecutionContext(
            self._pipeline_context_data,
            DagsterLogManager(
                self.run_id, merge_dicts(self.logging_tags, step.logging_tags), self.log.loggers
            ),
            step,
        )

    @property
    def executor_config(self):
        return self.run_config.executor_config

    @property
    def run_config(self):
        return self._pipeline_context_data.run_config

    @property
    def scoped_resources_builder(self):
        return self._pipeline_context_data.scoped_resources_builder

    @property
    def run_id(self):
        return self._pipeline_context_data.run_id

    @property
    def environment_dict(self):
        return self._pipeline_context_data.environment_dict

    @property
    def environment_config(self):
        return self._pipeline_context_data.environment_config

    @property
    def logging_tags(self):
        return self._log_manager.logging_tags

    def has_tag(self, key):
        check.str_param(key, 'key')
        return key in self.logging_tags

    def get_tag(self, key):
        check.str_param(key, 'key')
        return self.logging_tags[key]

    @property
    def pipeline_def(self):
        return self._pipeline_context_data.pipeline_def

    @property
    def mode_def(self):
        return self._pipeline_context_data.mode_def

    @property
    def system_storage_def(self):
        return self._pipeline_context_data.system_storage_def

    @property
    def event_callback(self):
        return self._pipeline_context_data.event_callback

    def has_event_callback(self):
        return self._pipeline_context_data.event_callback is not None

    @property
    def log(self):
        return self._log_manager

    @property
    def run_storage(self):
        return self._pipeline_context_data.run_storage

    @property
    def intermediates_manager(self):
        return self._pipeline_context_data.intermediates_manager

    @property
    def file_manager(self):
        return self._pipeline_context_data.file_manager


class SystemStepExecutionContext(SystemPipelineExecutionContext):
    __slots__ = ['_step', '_resources']

    def __init__(self, pipeline_context_data, log_manager, step):
        from dagster.core.execution.plan.objects import ExecutionStep

        self._step = check.inst_param(step, 'step', ExecutionStep)
        super(SystemStepExecutionContext, self).__init__(pipeline_context_data, log_manager)
        self._resources = self._pipeline_context_data.scoped_resources_builder.build(
            self.solid.resource_mapper_fn, self.solid_def.required_resource_keys
        )

    def for_transform(self):
        return SystemComputeExecutionContext(self._pipeline_context_data, self.log, self.step)

    def for_expectation(self, inout_def, expectation_def):
        return SystemExpectationExecutionContext(
            self._pipeline_context_data, self.log, self.step, inout_def, expectation_def
        )

    @property
    def step(self):
        return self._step

    @property
    def solid_handle(self):
        return self._step.solid_handle

    @property
    def solid_def(self):
        return self.solid.definition

    @property
    def solid(self):
        return self.pipeline_def.get_solid(self._step.solid_handle)

    @property
    def resources(self):
        return self._resources


class SystemComputeExecutionContext(SystemStepExecutionContext):
    @property
    def solid_config(self):
        solid_config = self.environment_config.solids.get(str(self.solid_handle))
        return solid_config.config if solid_config else None


class SystemExpectationExecutionContext(SystemStepExecutionContext):
    __slots__ = ['_inout_def', '_expectation_def']

    def __init__(self, pipeline_context_data, log_manager, step, inout_def, expectation_def):
        self._expectation_def = check.inst_param(
            expectation_def, 'expectation_def', IOExpectationDefinition
        )
        self._inout_def = check.inst_param(
            inout_def, 'inout_def', (InputDefinition, OutputDefinition)
        )
        super(SystemExpectationExecutionContext, self).__init__(
            pipeline_context_data, log_manager, step
        )

    @property
    def expectation_def(self):
        return self._expectation_def

    @property
    def inout_def(self):
        return self._inout_def



from collections import namedtuple
from contextlib import contextmanager
import inspect
import sys
import time

from contextlib2 import ExitStack

from dagster import check
from dagster.core.definitions import PipelineDefinition, create_environment_type
from dagster.core.definitions.handle import ExecutionTargetHandle
from dagster.core.definitions.resource import ScopedResourcesBuilder
from dagster.core.definitions.system_storage import SystemStorageData
from dagster.core.errors import (
    DagsterError,
    DagsterInvariantViolationError,
    DagsterUserCodeExecutionError,
    DagsterResourceFunctionError,
    DagsterInvalidConfigError,
    user_code_error_boundary,
)
from dagster.core.events import DagsterEvent, PipelineInitFailureData
from dagster.core.events.log import construct_event_logger
from dagster.core.log_manager import DagsterLogManager
from dagster.core.storage.init import InitSystemStorageContext
from dagster.core.storage.runs import DagsterRunMeta
from dagster.core.storage.type_storage import construct_type_storage_plugin_registry
from dagster.core.system_config.objects import EnvironmentConfig
from dagster.core.types.evaluator import evaluate_config
from dagster.loggers import default_loggers, default_system_loggers
from dagster.utils import merge_dicts
from dagster.utils.error import serializable_error_info_from_exc_info

from .config import RunConfig
from .context.init import InitResourceContext
from .context.system import SystemPipelineExecutionContextData, SystemPipelineExecutionContext
from .context.logger import InitLoggerContext


def create_environment_config(pipeline, environment_dict=None, run_config=None):
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.opt_dict_param(environment_dict, 'environment')
    run_config = check.opt_inst_param(run_config, 'run_config', RunConfig, default=RunConfig())

    mode = run_config.mode or pipeline.get_default_mode_name()
    environment_type = create_environment_type(pipeline, mode)

    result = evaluate_config(environment_type, environment_dict, pipeline, run_config)

    if not result.success:
        raise DagsterInvalidConfigError(pipeline, result.errors, environment_dict)

    return EnvironmentConfig.from_dict(result.value)


@contextmanager
def create_resource_builder(pipeline_def, environment_config, run_config, log_manager):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.inst_param(environment_config, 'environment_config', EnvironmentConfig)
    check.inst_param(run_config, 'run_config', RunConfig)
    check.inst_param(log_manager, 'log_manager', DagsterLogManager)

    resources_stack = ResourcesStack(pipeline_def, environment_config, run_config, log_manager)
    yield resources_stack.create()
    resources_stack.teardown()


def construct_system_storage_data(storage_init_context):
    return storage_init_context.system_storage_def.system_storage_creation_fn(storage_init_context)


def system_storage_def_from_config(mode_definition, environment_config):
    for system_storage_def in mode_definition.system_storage_defs:
        if system_storage_def.name == environment_config.storage.system_storage_name:
            return system_storage_def

    check.failed(
        'Could not find storage mode {}. Should have be caught by config system'.format(
            environment_config.storage.system_storage_name
        )
    )


def check_persistent_storage_requirement(pipeline_def, system_storage_def, run_config):
    if (
        run_config.executor_config.requires_persistent_storage
        and not system_storage_def.is_persistent
    ):
        raise DagsterInvariantViolationError(
            (
                'While invoking pipeline {pipeline_name}. You have attempted '
                'to use the multiprocessing executor while using system '
                'storage {storage_name} which does not persist intermediates. '
                'This means there would be no way to move data between different '
                'processes. Please configure your pipeline in the storage config '
                'section to use persistent system storage such as the filesystem.'
            ).format(pipeline_name=pipeline_def.name, storage_name=system_storage_def.name)
        )


# This represents all the data that is passed *into* context creation process.
# The remainder of the objects generated (e.g. loggers, resources) are created
# using user-defined code that may fail at runtime and result in the emission
# of a pipeline init failure events. The data in this object are passed all
# over the place during the context creation process so grouping here for
# ease of argument passing etc.
ContextCreationData = namedtuple(
    'ContextCreationData',
    'pipeline_def environment_config run_config mode_def system_storage_def execution_target_handle',
)


def create_context_creation_data(pipeline_def, environment_dict, run_config):
    environment_config = create_environment_config(pipeline_def, environment_dict, run_config)

    mode_def = pipeline_def.get_mode_definition(run_config.mode)
    system_storage_def = system_storage_def_from_config(mode_def, environment_config)

    check_persistent_storage_requirement(pipeline_def, system_storage_def, run_config)

    return ContextCreationData(
        pipeline_def=pipeline_def,
        environment_config=environment_config,
        run_config=run_config,
        mode_def=mode_def,
        system_storage_def=system_storage_def,
        execution_target_handle=ExecutionTargetHandle.get_handle(pipeline_def),
    )


@contextmanager
def scoped_pipeline_context(
    pipeline_def,
    environment_dict,
    run_config,
    system_storage_data=None,
    scoped_resources_builder_cm=create_resource_builder,
):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.dict_param(environment_dict, 'environment_dict', key_type=str)
    check.inst_param(run_config, 'run_config', RunConfig)
    check.opt_inst_param(system_storage_data, 'system_storage_data', SystemStorageData)

    context_creation_data = create_context_creation_data(pipeline_def, environment_dict, run_config)

    # After this try block, a Dagster exception thrown will result in a pipeline init failure event.
    try:
        log_manager = create_log_manager(context_creation_data)

        with scoped_resources_builder_cm(
            context_creation_data.pipeline_def,
            context_creation_data.environment_config,
            context_creation_data.run_config,
            log_manager,
        ) as scoped_resources_builder:

            system_storage_data = create_system_storage_data(
                context_creation_data, system_storage_data, scoped_resources_builder
            )

            yield construct_pipeline_execution_context(
                context_creation_data=context_creation_data,
                scoped_resources_builder=scoped_resources_builder,
                system_storage_data=system_storage_data,
                log_manager=log_manager,
            )

    except DagsterError as dagster_error:
        user_facing_exc_info = (
            # pylint does not know original_exc_info exists is is_user_code_error is true
            # pylint: disable=no-member
            dagster_error.original_exc_info
            if dagster_error.is_user_code_error
            else sys.exc_info()
        )

        if run_config.executor_config.raise_on_error:
            raise dagster_error

        error_info = serializable_error_info_from_exc_info(user_facing_exc_info)
        yield DagsterEvent.pipeline_init_failure(
            pipeline_name=pipeline_def.name,
            failure_data=PipelineInitFailureData(error=error_info),
            log_manager=_create_context_free_log_manager(run_config, pipeline_def),
        )


def create_system_storage_data(
    context_creation_data, system_storage_data, scoped_resources_builder
):
    check.inst_param(context_creation_data, 'context_creation_data', ContextCreationData)

    environment_config, pipeline_def, system_storage_def, run_config = (
        context_creation_data.environment_config,
        context_creation_data.pipeline_def,
        context_creation_data.system_storage_def,
        context_creation_data.run_config,
    )

    system_storage_data = (
        system_storage_data
        if system_storage_data
        else construct_system_storage_data(
            InitSystemStorageContext(
                pipeline_def=pipeline_def,
                mode_def=context_creation_data.mode_def,
                system_storage_def=system_storage_def,
                system_storage_config=environment_config.storage.system_storage_config,
                run_config=run_config,
                environment_config=environment_config,
                type_storage_plugin_registry=construct_type_storage_plugin_registry(
                    pipeline_def, system_storage_def
                ),
                resources=scoped_resources_builder.build(
                    # currently provide default for resource mapping
                    lambda resources, resources_deps: {r: resources.get(r) for r in resources_deps},
                    context_creation_data.system_storage_def.required_resource_keys,
                ),
            )
        )
    )

    system_storage_data.run_storage.write_dagster_run_meta(
        DagsterRunMeta(
            run_id=run_config.run_id, timestamp=time.time(), pipeline_name=pipeline_def.name
        )
    )
    return system_storage_data


def construct_pipeline_execution_context(
    context_creation_data, scoped_resources_builder, system_storage_data, log_manager
):
    check.inst_param(context_creation_data, 'context_creation_data', ContextCreationData)
    scoped_resources_builder = check.inst_param(
        scoped_resources_builder if scoped_resources_builder else ScopedResourcesBuilder(),
        'scoped_resources_builder',
        ScopedResourcesBuilder,
    )
    check.inst_param(system_storage_data, 'system_storage_data', SystemStorageData)
    check.inst_param(log_manager, 'log_manager', DagsterLogManager)

    return SystemPipelineExecutionContext(
        SystemPipelineExecutionContextData(
            pipeline_def=context_creation_data.pipeline_def,
            mode_def=context_creation_data.mode_def,
            system_storage_def=context_creation_data.system_storage_def,
            run_config=context_creation_data.run_config,
            scoped_resources_builder=scoped_resources_builder,
            environment_config=context_creation_data.environment_config,
            run_storage=system_storage_data.run_storage,
            intermediates_manager=system_storage_data.intermediates_manager,
            file_manager=system_storage_data.file_manager,
            execution_target_handle=context_creation_data.execution_target_handle,
        ),
        log_manager=log_manager,
    )


class ResourcesStack(object):
    # Wraps an ExitStack to support execution in environments like Dagstermill, where we can't
    # wrap solid execution/the pipeline execution context lifecycle in an ordinary Python context
    # manager (because notebook execution is cell-by-cell in the Jupyter kernel, a subprocess we
    # don't directly control). In these environments we need to manually create and teardown
    # resources.
    def __init__(self, pipeline_def, environment_config, run_config, log_manager):
        check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
        check.inst_param(environment_config, 'environment_config', EnvironmentConfig)
        check.inst_param(run_config, 'run_config', RunConfig)
        check.inst_param(log_manager, 'log_manager', DagsterLogManager)

        self.resource_instances = {}
        self.mode_definition = pipeline_def.get_mode_definition(run_config.mode)
        self.pipeline_def = pipeline_def
        self.environment_config = environment_config
        self.run_config = run_config
        self.log_manager = log_manager
        self.stack = ExitStack()

    def create(self):
        for resource_name, resource_def in sorted(self.mode_definition.resource_defs.items()):
            user_fn = create_resource_fn_lambda(
                self.pipeline_def,
                resource_def,
                self.environment_config.resources.get(resource_name, {}).get('config'),
                self.run_config.run_id,
                self.log_manager,
            )

            def _create_msg_fn(rn):
                return lambda: 'Error executing resource_fn on ResourceDefinition {name}'.format(
                    name=rn
                )

            resource_obj = self.stack.enter_context(
                user_code_context_manager(
                    user_fn, DagsterResourceFunctionError, _create_msg_fn(resource_name)
                )
            )

            self.resource_instances[resource_name] = resource_obj
        return ScopedResourcesBuilder(self.resource_instances)

    def teardown(self):
        self.stack.close()


def create_resource_fn_lambda(pipeline_def, resource_def, resource_config, run_id, log_manager):
    return lambda: resource_def.resource_fn(
        InitResourceContext(
            pipeline_def=pipeline_def,
            resource_def=resource_def,
            resource_config=resource_config,
            run_id=run_id,
            log_manager=log_manager,
        )
    )


def create_log_manager(context_creation_data):
    check.inst_param(context_creation_data, 'context_creation_data', ContextCreationData)

    pipeline_def, mode_def, environment_config, run_config = (
        context_creation_data.pipeline_def,
        context_creation_data.mode_def,
        context_creation_data.environment_config,
        context_creation_data.run_config,
    )

    loggers = []
    for logger_key, logger_def in mode_def.loggers.items() or default_loggers().items():
        if logger_key in environment_config.loggers:
            loggers.append(
                logger_def.logger_fn(
                    InitLoggerContext(
                        environment_config.loggers.get(logger_key, {}).get('config'),
                        pipeline_def,
                        logger_def,
                        run_config.run_id,
                    )
                )
            )

    if run_config.loggers:
        for logger in run_config.loggers:
            loggers.append(logger)

    if not loggers:
        for (logger_def, logger_config) in default_system_loggers():
            loggers.append(
                logger_def.logger_fn(
                    InitLoggerContext(logger_config, pipeline_def, logger_def, run_config.run_id)
                )
            )

    if run_config.event_callback:
        init_logger_context = InitLoggerContext({}, pipeline_def, logger_def, run_config.run_id)
        loggers.append(
            construct_event_logger(run_config.event_callback).logger_fn(init_logger_context)
        )

    return DagsterLogManager(
        run_id=run_config.run_id,
        logging_tags=get_logging_tags(
            context_creation_data.run_config, context_creation_data.pipeline_def
        ),
        loggers=loggers,
    )


def _create_context_free_log_manager(run_config, pipeline_def):
    '''In the event of pipeline initialization failure, we want to be able to log the failure
    without a dependency on the ExecutionContext to initialize DagsterLogManager.
    Args:
        run_config (dagster.core.execution_context.RunConfig)
        pipeline_def (dagster.definitions.PipelineDefinition)
    '''
    check.inst_param(run_config, 'run_config', RunConfig)
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)

    loggers = []
    # Use the default logger
    for (logger_def, logger_config) in default_system_loggers():
        loggers += [
            logger_def.logger_fn(
                InitLoggerContext(logger_config, pipeline_def, logger_def, run_config.run_id)
            )
        ]
    if run_config.event_callback:
        event_logger_def = construct_event_logger(run_config.event_callback)
        loggers += [
            event_logger_def.logger_fn(
                InitLoggerContext({}, pipeline_def, event_logger_def, run_config.run_id)
            )
        ]
    elif run_config.loggers:
        loggers += run_config.loggers

    return DagsterLogManager(run_config.run_id, get_logging_tags(run_config, pipeline_def), loggers)


def _ensure_gen(thing_or_gen):
    if not inspect.isgenerator(thing_or_gen):

        def _gen_thing():
            yield thing_or_gen

        return _gen_thing()

    return thing_or_gen


@contextmanager
def user_code_context_manager(user_fn, error_cls, msg_fn):
    '''Wraps the output of a user provided function that may yield or return a value and
    returns a generator that asserts it only yields a single value.
    '''
    check.callable_param(user_fn, 'user_fn')
    check.subclass_param(error_cls, 'error_cls', DagsterUserCodeExecutionError)

    with user_code_error_boundary(error_cls, msg_fn):
        thing_or_gen = user_fn()
        gen = _ensure_gen(thing_or_gen)

        try:
            thing = next(gen)
        except StopIteration:
            check.failed('Must yield one item. You did not yield anything.')

        yield thing

        stopped = False

        try:
            next(gen)
        except StopIteration:
            stopped = True

        check.invariant(stopped, 'Must yield one item. Yielded more than one item')


def get_logging_tags(run_config, pipeline):
    check.opt_inst_param(run_config, 'run_config', RunConfig)
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)

    return merge_dicts({'pipeline': pipeline.name}, run_config.tags if run_config else {})



from collections import namedtuple

from dagster import check

from dagster.core.definitions.logger import LoggerDefinition
from dagster.core.definitions.pipeline import PipelineDefinition


class InitLoggerContext(
    namedtuple('InitLoggerContext', 'logger_config pipeline_def logger_def run_id')
):
    '''
    Logger-specific initialization context.

    Attributes:
        logger_config (Any): The configuration data provided by the environment config. The
            schema for this data is defined by ``config_field`` on the :py:class:`LoggerDefinition`
        pipeline_def (PipelineDefinition): The pipeline definition currently being executed.
        logger_def (LoggerDefinition): The logger definition for the logger being constructed.
        run_id (str): The ID for this run of the pipeline.
    '''

    def __new__(cls, logger_config, pipeline_def, logger_def, run_id):
        return super(InitLoggerContext, cls).__new__(
            cls,
            logger_config,
            check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition),
            check.inst_param(logger_def, 'logger_def', LoggerDefinition),
            check.str_param(run_id, 'run_id'),
        )



from dagster import check
from dagster.core.errors import DagsterInvalidDefinitionError, DagsterInvariantViolationError

from .pipeline import PipelineDefinition


class RepositoryDefinition(object):
    '''Define a repository that contains a collection of pipelines.

    Args:
        name (str): The name of the pipeline.
        pipeline_dict (Dict[str, callable]):
            An dictionary of pipelines. The value of the dictionary is a function that takes
            no parameters and returns a PipelineDefiniton.

            We pass callables instead of the PipelineDefinitions itself so that they can be
            created on demand when accessed by name.

            As the pipelines are retrieved it ensures that the keys of the dictionary and the
            name of the pipeline are the same.
        pipeline_defs (List[PipelineDefinition]):
            A list of instantiated pipeline definitions.
    '''

    def __init__(self, name, pipeline_dict=None, pipeline_defs=None):
        self.name = check.str_param(name, 'name')

        pipeline_dict = check.opt_dict_param(pipeline_dict, 'pipeline_dict', key_type=str)
        pipeline_defs = check.opt_list_param(pipeline_defs, 'pipeline_defs', PipelineDefinition)

        for val in pipeline_dict.values():
            check.is_callable(val, 'Value in pipeline_dict must be function')

        self._lazy_pipeline_dict = pipeline_dict

        self._pipeline_cache = {}
        self._pipeline_names = set(pipeline_dict.keys())
        for defn in pipeline_defs:
            check.invariant(
                defn.name not in self._pipeline_names,
                'Duplicate pipelines named {name}'.format(name=defn.name),
            )
            self._pipeline_names.add(defn.name)
            self._pipeline_cache[defn.name] = defn

        self._all_pipelines = None
        self._solid_defs = None

    @property
    def pipeline_names(self):
        return self._pipeline_names

    def has_pipeline(self, name):
        check.str_param(name, 'name')
        return name in self._pipeline_names

    def get_pipeline(self, name):
        '''Get a pipeline by name. Only constructs that pipeline and caches it.

        Args:
            name (str): Name of the pipeline to retriever

        Returns:
            PipelineDefinition: Instance of PipelineDefinition with that name.
        '''
        check.str_param(name, 'name')

        if name in self._pipeline_cache:
            return self._pipeline_cache[name]

        try:
            pipeline = self._lazy_pipeline_dict[name]()
        except KeyError:
            raise DagsterInvariantViolationError(
                'Could not find pipeline "{name}". Found: {pipeline_names}.'.format(
                    name=name,
                    pipeline_names=', '.join(
                        [
                            '"{pipeline_name}"'.format(pipeline_name=pipeline_name)
                            for pipeline_name in self._pipeline_names
                        ]
                    ),
                )
            )
        check.invariant(
            pipeline.name == name,
            'Name does not match. Name in dict {name}. Name in pipeline {pipeline.name}'.format(
                name=name, pipeline=pipeline
            ),
        )

        self._pipeline_cache[name] = check.inst(
            pipeline,
            PipelineDefinition,
            (
                'Function passed into pipeline_dict with key {key} must return a '
                'PipelineDefinition'
            ).format(key=name),
        )

        return pipeline

    def get_all_pipelines(self):
        '''Return all pipelines as a list

        Returns:
            List[PipelineDefinition]:

        '''
        if self._all_pipelines:
            return self._all_pipelines

        self._all_pipelines = list(map(self.get_pipeline, self._pipeline_names))
        # This does uniqueness check
        self.get_all_solid_defs()
        return self._all_pipelines

    def get_all_solid_defs(self):
        if self._solid_defs:
            return self._solid_defs

        self._solid_defs = self._construct_solid_defs()
        return list(self._solid_defs.values())

    def _construct_solid_defs(self):
        solid_defs = {}
        solid_to_pipeline = {}
        # This looks like it should infinitely loop but the
        # memoization of all_pipelines and _solids_defs short
        # circuits that
        for pipeline in self.get_all_pipelines():
            for solid_def in pipeline.solid_defs:
                if solid_def.name not in solid_defs:
                    solid_defs[solid_def.name] = solid_def
                    solid_to_pipeline[solid_def.name] = pipeline.name

                if not solid_defs[solid_def.name] is solid_def:
                    first_name, second_name = sorted(
                        [solid_to_pipeline[solid_def.name], pipeline.name]
                    )
                    raise DagsterInvalidDefinitionError(
                        (
                            'You have defined two solid definitions named "{solid_def_name}" '
                            'in repository "{repository_name}". Solid definition names must be '
                            'unique within a repository. The solid definition has been defined in '
                            'pipeline "{first_pipeline_name}" and it has been defined '
                            'again in pipeline "{second_pipeline_name}."'
                        ).format(
                            solid_def_name=solid_def.name,
                            repository_name=self.name,
                            first_pipeline_name=first_name,
                            second_pipeline_name=second_name,
                        )
                    )

        return solid_defs

    def solid_def_named(self, name):
        check.str_param(name, 'name')

        self.get_all_solid_defs()

        if name not in self._solid_defs:
            check.failed('could not find solid_def {}'.format(name))

        return self._solid_defs[name]



from dagster.core.types import Int, Optional, List
from dagster.core.types.evaluator import evaluate_config
from dagster.core.types.field import resolve_to_config_type


def test_config_int():
    int_inst = resolve_to_config_type(Int)
    assert evaluate_config(int_inst, 1).success
    assert not evaluate_config(int_inst, None).success
    assert not evaluate_config(int_inst, 'r').success
    assert not int_inst.is_list
    assert not int_inst.is_nullable
    assert not (int_inst.is_nullable or int_inst.is_list)


def test_optional_int():
    optional_int_inst = resolve_to_config_type(Optional[Int])

    assert evaluate_config(optional_int_inst, 1).success
    assert evaluate_config(optional_int_inst, None).success
    assert not evaluate_config(optional_int_inst, 'r').success


def test_list_int():
    list_int = resolve_to_config_type(List[Int])

    assert evaluate_config(list_int, [1]).success
    assert evaluate_config(list_int, [1, 2]).success
    assert evaluate_config(list_int, []).success
    assert not evaluate_config(list_int, [None]).success
    assert not evaluate_config(list_int, [1, None]).success
    assert not evaluate_config(list_int, None).success
    assert not evaluate_config(list_int, [1, 'absdf']).success


def test_list_nullable_int():
    lni = resolve_to_config_type(List[Optional[Int]])

    assert evaluate_config(lni, [1]).success
    assert evaluate_config(lni, [1, 2]).success
    assert evaluate_config(lni, []).success
    assert evaluate_config(lni, [None]).success
    assert evaluate_config(lni, [1, None]).success
    assert not evaluate_config(lni, None).success
    assert not evaluate_config(lni, [1, 'absdf']).success



import sys
from dagster import check
from dagster.core.errors import DagsterInvalidDefinitionError


class WrappingType(object):
    def __init__(self, inner_type):
        # Cannot check inner_type because of circular references and no fwd declarations
        from dagster.core.types import BuiltinEnum

        if inner_type == BuiltinEnum.NOTHING:
            raise DagsterInvalidDefinitionError(
                'Type Nothing can not be wrapped in List or Optional'
            )

        self.inner_type = inner_type


class WrappingListType(WrappingType):
    pass


class WrappingNullableType(WrappingType):
    pass


if sys.version_info.major >= 3:
    import typing

    List = typing.List

    def is_python_list_typehint(type_annotation):
        origin = getattr(type_annotation, '__origin__', None)
        return origin == typing.List or origin == list

    def remap_to_dagster_list_type(type_annotation):
        check.invariant(
            is_python_list_typehint(type_annotation), 'type must pass is_python_list_typehint check'
        )
        return WrappingListType(type_annotation.__args__[0])

    Optional = typing.Optional

    def is_python_optional_typehint(type_annotation):
        # Optional[X] is Union[X, NoneType] which is what we match against here
        origin = getattr(type_annotation, '__origin__', None)
        return (
            origin == typing.Union
            and len(type_annotation.__args__) == 2
            and type_annotation.__args__[1] == type(None)
        )

    def remap_to_dagster_optional_type(type_annotation):
        check.invariant(
            is_python_optional_typehint(type_annotation),
            'type must pass is_python_optional_typehint check',
        )
        return WrappingNullableType(type_annotation.__args__[0])


else:

    class ListStub:
        def __getitem__(self, inner_type):
            return WrappingListType(inner_type)

    List = ListStub()

    def is_python_list_typehint(_):
        return False

    def remap_to_dagster_list_type(_):
        check.failed('Not valid to call in python 2')

    class OptionalStub:
        def __getitem__(self, inner_type):
            return WrappingNullableType(inner_type)

    Optional = OptionalStub()

    def is_python_optional_typehint(_):
        return False

    def remap_to_dagster_optional_type(_):
        check.failed('Not valid to call in python 2')
